"""Tests for MCP tools configuration loading and validation."""

import json
from collections.abc import Callable, Generator
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from deepagents_cli.mcp_tools import (
    MCPServerInfo,
    MCPSessionManager,
    MCPToolInfo,
    _check_remote_server,
    _check_stdio_server,
    _filter_project_stdio_servers,
    classify_discovered_configs,
    discover_mcp_configs,
    extract_stdio_server_commands,
    get_mcp_tools,
    load_mcp_config,
    load_mcp_config_lenient,
    merge_mcp_configs,
    resolve_and_load_mcp_tools,
)
from deepagents_cli.project_utils import ProjectContext

# Test Fixtures


@pytest.fixture
def valid_config_data() -> dict:
    """Fixture providing a valid stdio server configuration."""
    return {
        "mcpServers": {
            "filesystem": {
                "command": "npx",
                "args": ["-y", "@modelcontextprotocol/server-filesystem", "/tmp"],
                "env": {},
            }
        }
    }


@pytest.fixture
def write_config(tmp_path: Path) -> Callable[..., str]:
    """Fixture that writes a JSON config dict to a temp file and returns the path."""

    def _write(config_data: dict, filename: str = "mcp-config.json") -> str:
        config_file = tmp_path / filename
        config_file.write_text(json.dumps(config_data))
        return str(config_file)

    return _write


@pytest.fixture
def mock_mcp_session():
    """Fixture for creating a mock MCP session context manager."""
    mock_session = AsyncMock()
    mock_session_cm = MagicMock()
    mock_session_cm.__aenter__ = AsyncMock(return_value=mock_session)
    mock_session_cm.__aexit__ = AsyncMock(return_value=None)
    return mock_session, mock_session_cm


@pytest.fixture
def mock_mcp_client(
    mock_mcp_session: tuple[AsyncMock, MagicMock],
) -> tuple[MagicMock, AsyncMock]:
    """Fixture for creating a mock MultiServerMCPClient."""
    mock_session, mock_session_cm = mock_mcp_session
    mock_client = MagicMock()
    mock_client.session = MagicMock(return_value=mock_session_cm)
    return mock_client, mock_session


@pytest.fixture
def mock_tools():
    """Fixture providing mock tool objects."""
    mock_tool1 = MagicMock()
    mock_tool1.name = "read_file"
    mock_tool1.description = "Read a file"

    mock_tool2 = MagicMock()
    mock_tool2.name = "write_file"
    mock_tool2.description = "Write a file"

    return [mock_tool1, mock_tool2]


class TestLoadMCPConfig:
    """Test MCP configuration file loading and validation."""

    def test_load_valid_config(
        self, write_config: Callable[..., str], valid_config_data: dict
    ) -> None:
        """Test loading a valid MCP configuration file."""
        path = write_config(valid_config_data)
        config = load_mcp_config(path)
        assert config == valid_config_data

    def test_load_config_file_not_found(self, tmp_path: Path) -> None:
        """Test that FileNotFoundError is raised for missing config file."""
        nonexistent_file = tmp_path / "nonexistent.json"

        with pytest.raises(FileNotFoundError, match="MCP config file not found"):
            load_mcp_config(str(nonexistent_file))

    def test_load_config_invalid_json(self, tmp_path: Path) -> None:
        """Test that JSONDecodeError is raised for invalid JSON."""
        config_file = tmp_path / "invalid.json"
        config_file.write_text("{invalid json")

        with pytest.raises(
            json.JSONDecodeError, match="Invalid JSON in MCP config file"
        ):
            load_mcp_config(str(config_file))

    def test_load_config_missing_mcpservers_field(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test that ValueError is raised when mcpServers field is missing."""
        path = write_config({"someOtherField": "value"})

        with pytest.raises(ValueError, match="must contain 'mcpServers' field"):
            load_mcp_config(path)

    @pytest.mark.parametrize(
        ("config_data", "expected_error", "exception_type"),
        [
            (
                {"mcpServers": ["not", "a", "dict"]},
                "'mcpServers' field must be a dictionary",
                TypeError,
            ),
            ({"mcpServers": {}}, "'mcpServers' field is empty", ValueError),
        ],
    )
    def test_load_config_invalid_mcpservers(
        self,
        write_config: Callable[..., str],
        config_data: dict,
        expected_error: str,
        exception_type: type[Exception],
    ) -> None:
        """Test that appropriate exception is raised for invalid mcpServers field."""
        path = write_config(config_data)

        with pytest.raises(exception_type, match=expected_error):
            load_mcp_config(path)

    def test_load_config_server_missing_command(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test that ValueError is raised when server config is missing command."""
        path = write_config(
            {
                "mcpServers": {
                    "filesystem": {
                        "args": ["/tmp"],
                        # Missing "command" field
                    }
                }
            }
        )

        with pytest.raises(
            ValueError, match=r"filesystem.*missing required 'command' field"
        ):
            load_mcp_config(path)

    @pytest.mark.parametrize(
        ("server_config", "expected_error"),
        [
            ("not a dict", "filesystem.*config must be a dictionary"),
            (
                {"command": "npx", "args": "not a list"},
                "filesystem.*'args' must be a list",
            ),
            (
                {"command": "npx", "args": ["/tmp"], "env": ["not", "a", "dict"]},
                "filesystem.*'env' must be a dictionary",
            ),
        ],
    )
    def test_load_config_invalid_field_types(
        self,
        write_config: Callable[..., str],
        server_config: dict | str,
        expected_error: str,
    ) -> None:
        """Test that TypeError is raised for invalid server config field types."""
        path = write_config({"mcpServers": {"filesystem": server_config}})

        with pytest.raises(TypeError, match=expected_error):
            load_mcp_config(path)

    def test_load_config_optional_fields(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test that args and env are optional fields."""
        config_data = {
            "mcpServers": {
                "simple": {
                    "command": "simple-server",
                    # No args or env - should be valid
                }
            }
        }
        path = write_config(config_data)

        config = load_mcp_config(path)

        assert config == config_data
        assert "simple" in config["mcpServers"]

    def test_load_config_multiple_servers(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test loading config with multiple MCP servers."""
        config_data = {
            "mcpServers": {
                "filesystem": {
                    "command": "npx",
                    "args": ["-y", "@modelcontextprotocol/server-filesystem", "/tmp"],
                    "env": {},
                },
                "brave-search": {
                    "command": "npx",
                    "args": ["-y", "@modelcontextprotocol/server-brave-search"],
                    "env": {"BRAVE_API_KEY": "test-key"},
                },
                "github": {
                    "command": "npx",
                    "args": ["-y", "@modelcontextprotocol/server-github"],
                    "env": {"GITHUB_TOKEN": "test-token"},
                },
            }
        }
        path = write_config(config_data)

        config = load_mcp_config(path)

        assert len(config["mcpServers"]) == 3
        assert "filesystem" in config["mcpServers"]
        assert "brave-search" in config["mcpServers"]
        assert "github" in config["mcpServers"]

    def test_load_config_sse_server(self, write_config: Callable[..., str]) -> None:
        """Test loading config with SSE server type."""
        config_data = {
            "mcpServers": {
                "remote-api": {
                    "type": "sse",
                    "url": "https://api.example.com/mcp",
                }
            }
        }
        path = write_config(config_data)

        config = load_mcp_config(path)

        assert config == config_data
        assert config["mcpServers"]["remote-api"]["type"] == "sse"
        assert (
            config["mcpServers"]["remote-api"]["url"] == "https://api.example.com/mcp"
        )

    def test_load_config_http_server(self, write_config: Callable[..., str]) -> None:
        """Test loading config with HTTP server type."""
        config_data = {
            "mcpServers": {
                "web-api": {
                    "type": "http",
                    "url": "https://api.example.com/mcp",
                }
            }
        }
        path = write_config(config_data)

        config = load_mcp_config(path)

        assert config == config_data
        assert config["mcpServers"]["web-api"]["type"] == "http"

    @pytest.mark.parametrize(
        ("server_name", "server_type"),
        [
            ("remote-api", "sse"),
            ("web-api", "http"),
        ],
    )
    def test_load_config_remote_server_missing_url(
        self,
        write_config: Callable[..., str],
        server_name: str,
        server_type: str,
    ) -> None:
        """Test that ValueError is raised when SSE/HTTP server is missing url field."""
        path = write_config({"mcpServers": {server_name: {"type": server_type}}})

        with pytest.raises(
            ValueError, match=f"{server_name}.*missing required 'url' field"
        ):
            load_mcp_config(path)

    def test_load_config_mixed_server_types(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test loading config with mixed stdio, SSE, and HTTP servers."""
        config_data = {
            "mcpServers": {
                "filesystem": {
                    "command": "npx",
                    "args": ["-y", "@modelcontextprotocol/server-filesystem", "/tmp"],
                },
                "remote-sse": {
                    "type": "sse",
                    "url": "https://api.example.com/sse",
                },
                "remote-http": {
                    "type": "http",
                    "url": "https://api.example.com/http",
                },
            }
        }
        path = write_config(config_data)

        config = load_mcp_config(path)

        assert len(config["mcpServers"]) == 3
        assert "command" in config["mcpServers"]["filesystem"]
        assert config["mcpServers"]["remote-sse"]["type"] == "sse"
        assert config["mcpServers"]["remote-http"]["type"] == "http"

    def test_load_config_sse_with_headers(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test loading SSE server config with custom headers."""
        config_data = {
            "mcpServers": {
                "authenticated-api": {
                    "type": "sse",
                    "url": "https://api.example.com/mcp",
                    "headers": {
                        "Authorization": "Bearer token123",
                        "X-Custom-Header": "value",
                    },
                }
            }
        }
        path = write_config(config_data)

        config = load_mcp_config(path)

        headers = config["mcpServers"]["authenticated-api"]["headers"]
        assert headers["Authorization"] == "Bearer token123"

    def test_load_config_http_with_headers(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test loading HTTP server config with custom headers."""
        config_data = {
            "mcpServers": {
                "authenticated-api": {
                    "type": "http",
                    "url": "https://api.example.com/mcp",
                    "headers": {"Authorization": "Bearer secret"},
                }
            }
        }
        path = write_config(config_data)

        config = load_mcp_config(path)

        headers = config["mcpServers"]["authenticated-api"]["headers"]
        assert headers["Authorization"] == "Bearer secret"

    @pytest.mark.parametrize(
        ("transport_field", "transport_value"),
        [
            ("transport", "http"),
            ("transport", "sse"),
            ("type", "http"),
            ("type", "sse"),
        ],
    )
    def test_load_config_transport_field_alias(
        self,
        write_config: Callable[..., str],
        transport_field: str,
        transport_value: str,
    ) -> None:
        """Test that both 'type' and 'transport' fields are accepted for server type."""
        config_data = {
            "mcpServers": {
                "remote": {
                    transport_field: transport_value,
                    "url": "https://api.example.com/mcp",
                }
            }
        }
        path = write_config(config_data)

        config = load_mcp_config(path)

        assert config == config_data

    def test_load_config_invalid_headers_type(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test that TypeError is raised when headers is not a dictionary."""
        path = write_config(
            {
                "mcpServers": {
                    "api": {
                        "type": "sse",
                        "url": "https://api.example.com/mcp",
                        "headers": ["not", "a", "dict"],
                    }
                }
            }
        )

        with pytest.raises(TypeError, match=r"api.*'headers' must be a dictionary"):
            load_mcp_config(path)

    def test_load_config_unknown_server_type(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test that ValueError is raised for unsupported transport types."""
        path = write_config(
            {
                "mcpServers": {
                    "ws-server": {
                        "type": "websocket",
                        "url": "ws://example.com/mcp",
                    }
                }
            }
        )

        with pytest.raises(
            ValueError, match=r"ws-server.*unsupported transport type 'websocket'"
        ):
            load_mcp_config(path)


class TestGetMCPTools:
    """Test MCP tools loading from configuration."""

    @pytest.fixture(autouse=True)
    def _bypass_health_checks(self) -> Generator[None]:
        """Bypass pre-flight health checks for all tests in this class."""
        with (
            patch("deepagents_cli.mcp_tools._check_stdio_server"),
            patch(
                "deepagents_cli.mcp_tools._check_remote_server",
                new_callable=AsyncMock,
            ),
        ):
            yield

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_success(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
        valid_config_data: dict,
        mock_mcp_client: tuple,
        mock_tools: list,
    ) -> None:
        """Test successful loading of MCP tools."""
        path = write_config(valid_config_data)

        # Setup mocks
        mock_client, mock_session = mock_mcp_client
        mock_client_class.return_value = mock_client
        mock_load_tools.return_value = mock_tools

        tools, manager, server_infos = await get_mcp_tools(path)

        # Verify client was initialized with correct connection config
        mock_client_class.assert_called_once()
        connections = mock_client_class.call_args.kwargs["connections"]
        assert connections["filesystem"]["command"] == "npx"
        assert connections["filesystem"]["transport"] == "stdio"
        assert connections["filesystem"]["args"] == [
            "-y",
            "@modelcontextprotocol/server-filesystem",
            "/tmp",
        ]

        # Verify session was created and tools were loaded
        mock_client.session.assert_called_once_with("filesystem")
        mock_load_tools.assert_called_once_with(
            mock_session, server_name="filesystem", tool_name_prefix=True
        )
        assert len(tools) == 2
        assert tools[0].name == "read_file"
        assert tools[1].name == "write_file"
        assert isinstance(manager, MCPSessionManager)

        # Verify server_infos
        assert len(server_infos) == 1
        assert server_infos[0].name == "filesystem"
        assert server_infos[0].transport == "stdio"
        assert len(server_infos[0].tools) == 2
        assert server_infos[0].tools[0] == MCPToolInfo(
            name="read_file", description="Read a file"
        )
        assert server_infos[0].tools[1] == MCPToolInfo(
            name="write_file", description="Write a file"
        )

        # Clean up
        await manager.cleanup()

    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_server_spawn_failure(
        self, mock_client_class: MagicMock, write_config: Callable[..., str]
    ) -> None:
        """Test handling of MCP server spawn failure."""
        path = write_config(
            {
                "mcpServers": {
                    "filesystem": {
                        "command": "nonexistent-command",
                        "args": [],
                        "env": {},
                    }
                }
            }
        )

        # Setup mock client to raise an exception
        mock_client_class.side_effect = Exception("Command not found")

        with pytest.raises(
            RuntimeError, match=r"Failed to initialize MCP client.*Command not found"
        ):
            await get_mcp_tools(path)

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_get_tools_failure(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
        valid_config_data: dict,
        mock_mcp_client: tuple,
    ) -> None:
        """Test handling of failure during load_mcp_tools call."""
        path = write_config(valid_config_data)

        # Setup mocks
        mock_client, _ = mock_mcp_client
        mock_client_class.return_value = mock_client
        mock_load_tools.side_effect = Exception("Server protocol error")

        with pytest.raises(
            RuntimeError,
            match=r"Failed to load tools from MCP server.*Server protocol error",
        ):
            await get_mcp_tools(path)

    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_cleanup_called_on_client_init_failure(
        self, mock_client_class: MagicMock, write_config: Callable[..., str]
    ) -> None:
        """Test that manager.cleanup() is called when client init fails."""
        path = write_config(
            {"mcpServers": {"fs": {"command": "npx", "args": [], "env": {}}}}
        )
        mock_client_class.side_effect = Exception("init boom")

        with patch.object(
            MCPSessionManager, "cleanup", new_callable=AsyncMock
        ) as mock_cleanup:
            with pytest.raises(RuntimeError, match="Failed to initialize MCP client"):
                await get_mcp_tools(path)
            mock_cleanup.assert_awaited_once()

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_cleanup_called_on_tool_load_failure(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
        valid_config_data: dict,
        mock_mcp_client: tuple,
    ) -> None:
        """Test that manager.cleanup() is called when tool loading fails."""
        path = write_config(valid_config_data)
        mock_client, _ = mock_mcp_client
        mock_client_class.return_value = mock_client
        mock_load_tools.side_effect = Exception("tool boom")

        with patch.object(
            MCPSessionManager, "cleanup", new_callable=AsyncMock
        ) as mock_cleanup:
            with pytest.raises(RuntimeError, match="Failed to load tools"):
                await get_mcp_tools(path)
            mock_cleanup.assert_awaited_once()

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_empty_env_dict_coerced_to_none(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
        mock_mcp_client: tuple,
    ) -> None:
        """Test that `"env": {}` is coerced to None (inherit parent env)."""
        path = write_config(
            {"mcpServers": {"fs": {"command": "npx", "args": [], "env": {}}}}
        )
        mock_client, _ = mock_mcp_client
        mock_client_class.return_value = mock_client
        mock_load_tools.return_value = []

        _, manager, _ = await get_mcp_tools(path)

        connections = mock_client_class.call_args.kwargs["connections"]
        assert connections["fs"]["env"] is None

        await manager.cleanup()

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_multiple_servers(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
    ) -> None:
        """Test loading tools from multiple MCP servers."""
        path = write_config(
            {
                "mcpServers": {
                    "filesystem": {
                        "command": "npx",
                        "args": [
                            "-y",
                            "@modelcontextprotocol/server-filesystem",
                            "/tmp",
                        ],
                        "env": {},
                    },
                    "brave-search": {
                        "command": "npx",
                        "args": ["-y", "@modelcontextprotocol/server-brave-search"],
                        "env": {"BRAVE_API_KEY": "test-key"},
                    },
                }
            }
        )

        # Create mock tools from different servers
        mock_tools_fs = [MagicMock(name="read_file"), MagicMock(name="write_file")]
        mock_tools_search = [MagicMock(name="web_search")]

        # Setup mock client with session support
        mock_session_fs = AsyncMock()
        mock_session_search = AsyncMock()

        # Mock session context managers for both servers
        def mock_session_cm(server_name: str) -> MagicMock:
            session = (
                mock_session_fs if server_name == "filesystem" else mock_session_search
            )
            cm = MagicMock()
            cm.__aenter__ = AsyncMock(return_value=session)
            cm.__aexit__ = AsyncMock(return_value=None)
            return cm

        mock_client = MagicMock()
        mock_client.session.side_effect = mock_session_cm
        mock_client_class.return_value = mock_client

        # Mock load_mcp_tools to return different tools for each session
        def mock_load_side_effect(
            session: AsyncMock, **_kwargs: object
        ) -> list[MagicMock]:
            if session == mock_session_fs:
                return mock_tools_fs
            return mock_tools_search

        mock_load_tools.side_effect = mock_load_side_effect

        tools, manager, server_infos = await get_mcp_tools(path)

        # Verify both servers were registered
        call_kwargs = mock_client_class.call_args.kwargs
        connections = call_kwargs["connections"]
        assert len(connections) == 2
        assert "filesystem" in connections
        assert "brave-search" in connections
        assert connections["brave-search"]["env"]["BRAVE_API_KEY"] == "test-key"

        # Verify sessions were created for both servers
        assert mock_client.session.call_count == 2

        # Verify tools from all servers were returned
        assert len(tools) == 3

        # Verify server_infos for multiple servers
        assert len(server_infos) == 2
        assert server_infos[0].name == "filesystem"
        assert server_infos[0].transport == "stdio"
        assert len(server_infos[0].tools) == 2
        assert server_infos[1].name == "brave-search"
        assert server_infos[1].transport == "stdio"
        assert len(server_infos[1].tools) == 1

        # Clean up
        await manager.cleanup()

    async def test_get_mcp_tools_invalid_config(
        self, write_config: Callable[..., str]
    ) -> None:
        """Test that config validation errors are propagated."""
        path = write_config(
            {
                "mcpServers": {
                    "filesystem": {
                        "args": ["/tmp"],
                        # Missing command field
                    }
                }
            }
        )

        with pytest.raises(ValueError, match="missing required 'command' field"):
            await get_mcp_tools(path)

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_env_variables_passed(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
        mock_mcp_client: tuple,
    ) -> None:
        """Test that environment variables are correctly passed to MCP client."""
        path = write_config(
            {
                "mcpServers": {
                    "github": {
                        "command": "npx",
                        "args": ["-y", "@modelcontextprotocol/server-github"],
                        "env": {
                            "GITHUB_TOKEN": "ghp_test123",
                            "GITHUB_API_URL": "https://api.github.com",
                        },
                    }
                }
            }
        )

        # Setup mocks
        mock_client, _ = mock_mcp_client
        mock_client_class.return_value = mock_client
        mock_load_tools.return_value = []

        _, manager, _ = await get_mcp_tools(path)

        # Verify env variables were passed correctly
        connections = mock_client_class.call_args.kwargs["connections"]
        assert connections["github"]["env"]["GITHUB_TOKEN"] == "ghp_test123"
        assert (
            connections["github"]["env"]["GITHUB_API_URL"] == "https://api.github.com"
        )

        # Clean up
        await manager.cleanup()

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_env_none_when_not_provided(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
        mock_mcp_client: tuple,
    ) -> None:
        """Test that env is None (inherit parent env) when not provided in config."""
        path = write_config(
            {
                "mcpServers": {
                    "simple": {
                        "command": "simple-server",
                    }
                }
            }
        )

        # Setup mocks
        mock_client, _ = mock_mcp_client
        mock_client_class.return_value = mock_client
        mock_load_tools.return_value = []

        _, manager, _ = await get_mcp_tools(path)

        connections = mock_client_class.call_args.kwargs["connections"]
        assert connections["simple"]["env"] is None

        await manager.cleanup()

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_headers_passed_for_sse(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
        mock_mcp_client: tuple,
    ) -> None:
        """Test that headers are correctly passed to SSE MCP client."""
        path = write_config(
            {
                "mcpServers": {
                    "api": {
                        "type": "sse",
                        "url": "https://api.example.com/mcp",
                        "headers": {
                            "Authorization": "Bearer token123",
                            "X-API-Key": "key456",
                        },
                    }
                }
            }
        )

        # Setup mocks
        mock_client, _ = mock_mcp_client
        mock_client_class.return_value = mock_client
        mock_load_tools.return_value = []

        _, manager, _ = await get_mcp_tools(path)

        # Verify headers were passed correctly
        connections = mock_client_class.call_args.kwargs["connections"]
        assert connections["api"]["transport"] == "sse"
        assert connections["api"]["headers"]["Authorization"] == "Bearer token123"
        assert connections["api"]["headers"]["X-API-Key"] == "key456"

        # Clean up
        await manager.cleanup()

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_headers_passed_for_http(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
        mock_mcp_client: tuple,
    ) -> None:
        """Test that headers are correctly passed to HTTP MCP client."""
        path = write_config(
            {
                "mcpServers": {
                    "api": {
                        "type": "http",
                        "url": "https://api.example.com/mcp",
                        "headers": {"Authorization": "Bearer secret"},
                    }
                }
            }
        )

        # Setup mocks
        mock_client, _ = mock_mcp_client
        mock_client_class.return_value = mock_client
        mock_load_tools.return_value = []

        _, manager, _ = await get_mcp_tools(path)

        # Verify headers were passed and transport is correct
        connections = mock_client_class.call_args.kwargs["connections"]
        assert connections["api"]["transport"] == "streamable_http"
        assert connections["api"]["headers"]["Authorization"] == "Bearer secret"

        # Clean up
        await manager.cleanup()

    @patch("langchain_mcp_adapters.tools.load_mcp_tools")
    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_get_mcp_tools_no_headers_when_not_provided(
        self,
        mock_client_class: MagicMock,
        mock_load_tools: AsyncMock,
        write_config: Callable[..., str],
        mock_mcp_client: tuple,
    ) -> None:
        """Test that headers key is not added when not provided in config."""
        path = write_config(
            {
                "mcpServers": {
                    "api": {
                        "type": "sse",
                        "url": "https://api.example.com/mcp",
                    }
                }
            }
        )

        # Setup mocks
        mock_client, _ = mock_mcp_client
        mock_client_class.return_value = mock_client
        mock_load_tools.return_value = []

        _, manager, _ = await get_mcp_tools(path)

        # Verify headers key is not present
        connections = mock_client_class.call_args.kwargs["connections"]
        assert "headers" not in connections["api"]

        # Clean up
        await manager.cleanup()


class TestDiscoverMcpConfigs:
    """Test auto-discovery of MCP config files."""

    def test_project_context_overrides_process_cwd(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Explicit project context should drive discovery instead of cwd."""
        home = tmp_path / "home"
        home.mkdir()
        monkeypatch.setattr(Path, "home", lambda: home)

        project_root = tmp_path / "project"
        project_root.mkdir()
        (project_root / ".git").mkdir()
        user_cwd = project_root / "src"
        user_cwd.mkdir()
        project_cfg = project_root / ".mcp.json"
        project_cfg.write_text('{"mcpServers": {}}')

        other_cwd = tmp_path / "elsewhere"
        other_cwd.mkdir()
        monkeypatch.chdir(other_cwd)

        project_context = ProjectContext.from_user_cwd(user_cwd)
        result = discover_mcp_configs(project_context=project_context)

        assert result == [project_cfg]

    def test_no_configs_found(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Empty list when no config files exist."""
        project = tmp_path / "project"
        monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
        monkeypatch.setattr(
            "deepagents_cli.project_utils.find_project_root",
            lambda _start_path=None: project,
        )
        (tmp_path / "home").mkdir()
        project.mkdir()
        assert discover_mcp_configs() == []

    def test_user_level_only(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Only ~/.deepagents/.mcp.json exists."""
        home = tmp_path / "home"
        user_dir = home / ".deepagents"
        user_dir.mkdir(parents=True)
        cfg = user_dir / ".mcp.json"
        cfg.write_text('{"mcpServers": {}}')

        project = tmp_path / "project"
        project.mkdir()

        monkeypatch.setattr(Path, "home", lambda: home)
        monkeypatch.setattr(
            "deepagents_cli.project_utils.find_project_root",
            lambda _start_path=None: project,
        )
        result = discover_mcp_configs()
        assert result == [cfg]

    def test_project_root_only(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Only <project>/.mcp.json exists."""
        home = tmp_path / "home"
        home.mkdir()
        project = tmp_path / "project"
        project.mkdir()
        cfg = project / ".mcp.json"
        cfg.write_text('{"mcpServers": {}}')

        monkeypatch.setattr(Path, "home", lambda: home)
        monkeypatch.setattr(
            "deepagents_cli.project_utils.find_project_root",
            lambda _start_path=None: project,
        )
        result = discover_mcp_configs()
        assert result == [cfg]

    def test_project_deepagents_subdir_only(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Only <project>/.deepagents/.mcp.json exists."""
        home = tmp_path / "home"
        home.mkdir()
        project = tmp_path / "project"
        subdir = project / ".deepagents"
        subdir.mkdir(parents=True)
        cfg = subdir / ".mcp.json"
        cfg.write_text('{"mcpServers": {}}')

        monkeypatch.setattr(Path, "home", lambda: home)
        monkeypatch.setattr(
            "deepagents_cli.project_utils.find_project_root",
            lambda _start_path=None: project,
        )
        result = discover_mcp_configs()
        assert result == [cfg]

    def test_all_three_locations(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """All three config locations exist, returned in precedence order."""
        home = tmp_path / "home"
        user_dir = home / ".deepagents"
        user_dir.mkdir(parents=True)
        user_cfg = user_dir / ".mcp.json"
        user_cfg.write_text('{"mcpServers": {}}')

        project = tmp_path / "project"
        proj_subdir = project / ".deepagents"
        proj_subdir.mkdir(parents=True)
        proj_sub_cfg = proj_subdir / ".mcp.json"
        proj_sub_cfg.write_text('{"mcpServers": {}}')

        proj_root_cfg = project / ".mcp.json"
        proj_root_cfg.write_text('{"mcpServers": {}}')

        monkeypatch.setattr(Path, "home", lambda: home)
        monkeypatch.setattr(
            "deepagents_cli.project_utils.find_project_root",
            lambda _start_path=None: project,
        )
        result = discover_mcp_configs()
        assert result == [user_cfg, proj_sub_cfg, proj_root_cfg]

    def test_falls_back_to_cwd_when_no_git(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Falls back to CWD when find_project_root returns None."""
        home = tmp_path / "home"
        home.mkdir()
        monkeypatch.setattr(Path, "home", lambda: home)
        monkeypatch.setattr(
            "deepagents_cli.project_utils.find_project_root",
            lambda _start_path=None: None,
        )
        monkeypatch.chdir(tmp_path)
        cfg = tmp_path / ".mcp.json"
        cfg.write_text('{"mcpServers": {}}')

        result = discover_mcp_configs()
        assert cfg in result


class TestMergeMcpConfigs:
    """Test merging multiple MCP config dicts."""

    def test_single_config(self) -> None:
        """Single config passes through."""
        cfg = {"mcpServers": {"fs": {"command": "npx"}}}
        result = merge_mcp_configs([cfg])
        assert result == cfg

    def test_disjoint_servers(self) -> None:
        """Disjoint server names are all present."""
        c1 = {"mcpServers": {"fs": {"command": "npx"}}}
        c2 = {"mcpServers": {"search": {"command": "brave"}}}
        result = merge_mcp_configs([c1, c2])
        assert "fs" in result["mcpServers"]
        assert "search" in result["mcpServers"]

    def test_duplicate_server_name_last_wins(self) -> None:
        """Later config overrides earlier for same server name."""
        c1 = {"mcpServers": {"fs": {"command": "old"}}}
        c2 = {"mcpServers": {"fs": {"command": "new"}}}
        result = merge_mcp_configs([c1, c2])
        assert result["mcpServers"]["fs"]["command"] == "new"

    def test_empty_list(self) -> None:
        """Empty input returns empty mcpServers."""
        result = merge_mcp_configs([])
        assert result == {"mcpServers": {}}


class TestLoadMcpConfigLenient:
    """Test lenient config loading for auto-discovery."""

    def test_missing_file_returns_none(self, tmp_path: Path) -> None:
        """Missing file returns None without raising."""
        result = load_mcp_config_lenient(tmp_path / "nonexistent.json")
        assert result is None

    def test_invalid_json_returns_none(self, tmp_path: Path) -> None:
        """Invalid JSON returns None and logs warning."""
        bad = tmp_path / "bad.json"
        bad.write_text("{not json")
        result = load_mcp_config_lenient(bad)
        assert result is None

    def test_validation_error_returns_none(self, tmp_path: Path) -> None:
        """Config missing mcpServers returns None."""
        bad = tmp_path / "bad.json"
        bad.write_text('{"other": true}')
        result = load_mcp_config_lenient(bad)
        assert result is None

    def test_valid_file_returns_config(self, tmp_path: Path) -> None:
        """Valid config file returns parsed dict."""
        good = tmp_path / "good.json"
        good.write_text(
            json.dumps({"mcpServers": {"fs": {"command": "npx", "args": []}}})
        )
        result = load_mcp_config_lenient(good)
        assert result is not None
        assert "fs" in result["mcpServers"]


class TestResolveAndLoadMcpTools:
    """Test the unified resolve_and_load_mcp_tools entry point."""

    async def test_no_mcp_returns_empty(self) -> None:
        """no_mcp=True returns empty tuple immediately."""
        tools, manager, infos = await resolve_and_load_mcp_tools(no_mcp=True)
        assert tools == []
        assert manager is None
        assert infos == []

    @patch("deepagents_cli.mcp_tools._load_tools_from_config")
    @patch("deepagents_cli.mcp_tools.discover_mcp_configs")
    async def test_explicit_path_merges_with_discovery(
        self,
        mock_discover: MagicMock,
        mock_load: AsyncMock,
        tmp_path: Path,
    ) -> None:
        """Explicit path is merged on top of auto-discovered configs."""
        # Auto-discovered config
        discovered = tmp_path / "discovered.json"
        discovered.write_text(
            json.dumps({"mcpServers": {"fs": {"command": "npx", "args": []}}})
        )
        mock_discover.return_value = [discovered]

        # Explicit config
        explicit = tmp_path / "explicit.json"
        explicit.write_text(
            json.dumps({"mcpServers": {"search": {"command": "brave", "args": []}}})
        )
        mock_load.return_value = ([], MCPSessionManager(), [])

        await resolve_and_load_mcp_tools(
            explicit_config_path=str(explicit), trust_project_mcp=True
        )

        mock_discover.assert_called_once()
        mock_load.assert_awaited_once()
        merged = mock_load.call_args.args[0]
        assert "fs" in merged["mcpServers"]
        assert "search" in merged["mcpServers"]

    @patch("deepagents_cli.mcp_tools._load_tools_from_config")
    @patch("deepagents_cli.mcp_tools.discover_mcp_configs")
    async def test_auto_discovery_merges_and_loads(
        self, mock_discover: MagicMock, mock_load: AsyncMock, tmp_path: Path
    ) -> None:
        """Auto-discovery finds configs, merges, and loads tools."""
        # Write two config files
        c1 = tmp_path / "user.json"
        c1.write_text(
            json.dumps({"mcpServers": {"fs": {"command": "npx", "args": []}}})
        )
        c2 = tmp_path / "project.json"
        c2.write_text(
            json.dumps({"mcpServers": {"search": {"command": "brave", "args": []}}})
        )
        mock_discover.return_value = [c1, c2]
        mock_load.return_value = ([], MCPSessionManager(), [])

        await resolve_and_load_mcp_tools(trust_project_mcp=True)

        mock_load.assert_awaited_once()
        merged = mock_load.call_args.args[0]
        assert "fs" in merged["mcpServers"]
        assert "search" in merged["mcpServers"]

    @patch("deepagents_cli.mcp_tools.discover_mcp_configs")
    async def test_auto_discovery_no_configs_returns_empty(
        self, mock_discover: MagicMock
    ) -> None:
        """No discovered configs returns empty tuple."""
        mock_discover.return_value = []
        tools, manager, infos = await resolve_and_load_mcp_tools()
        assert tools == []
        assert manager is None
        assert infos == []

    async def test_explicit_path_missing_raises(self, tmp_path: Path) -> None:
        """FileNotFoundError propagates for missing explicit config."""
        with pytest.raises(FileNotFoundError):
            await resolve_and_load_mcp_tools(
                explicit_config_path=str(tmp_path / "nope.json")
            )

    async def test_explicit_path_invalid_json_raises(self, tmp_path: Path) -> None:
        """JSONDecodeError propagates for invalid explicit config."""
        bad = tmp_path / "bad.json"
        bad.write_text("{not json")
        with pytest.raises(json.JSONDecodeError):
            await resolve_and_load_mcp_tools(explicit_config_path=str(bad))

    @patch("deepagents_cli.mcp_tools.discover_mcp_configs")
    async def test_no_mcp_skips_discovery(self, mock_discover: MagicMock) -> None:
        """no_mcp=True should not call discover_mcp_configs."""
        await resolve_and_load_mcp_tools(no_mcp=True)
        mock_discover.assert_not_called()

    @patch("deepagents_cli.mcp_tools._load_tools_from_config")
    @patch("deepagents_cli.mcp_trust.is_project_mcp_trusted")
    async def test_project_context_drives_trust_root(
        self,
        mock_is_trusted: MagicMock,
        mock_load: AsyncMock,
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Trust lookup should use explicit project context, not process cwd."""
        home = tmp_path / "home"
        home.mkdir()
        monkeypatch.setattr(Path, "home", lambda: home)

        project_root = tmp_path / "project"
        project_root.mkdir()
        (project_root / ".git").mkdir()
        user_cwd = project_root / "src"
        user_cwd.mkdir()
        project_cfg = project_root / ".mcp.json"
        project_cfg.write_text(
            json.dumps({"mcpServers": {"local": {"command": "npx", "args": []}}})
        )

        other_cwd = tmp_path / "elsewhere"
        other_cwd.mkdir()
        monkeypatch.chdir(other_cwd)

        mock_is_trusted.return_value = True
        mock_load.return_value = ([], MCPSessionManager(), [])

        project_context = ProjectContext.from_user_cwd(user_cwd)
        await resolve_and_load_mcp_tools(project_context=project_context)

        mock_is_trusted.assert_called_once()
        assert mock_is_trusted.call_args.args[0] == str(project_root.resolve())
        mock_load.assert_awaited_once()

    @patch("deepagents_cli.mcp_tools._load_tools_from_config")
    async def test_project_context_normalizes_relative_explicit_path(
        self,
        mock_load: AsyncMock,
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Explicit config paths should resolve relative to project context cwd."""
        home = tmp_path / "home"
        home.mkdir()
        monkeypatch.setattr(Path, "home", lambda: home)

        project_root = tmp_path / "project"
        project_root.mkdir()
        (project_root / ".git").mkdir()
        user_cwd = project_root / "src"
        user_cwd.mkdir()
        explicit = user_cwd / "configs" / "mcp.json"
        explicit.parent.mkdir(parents=True)
        explicit.write_text(
            json.dumps({"mcpServers": {"fs": {"command": "npx", "args": []}}})
        )

        mock_load.return_value = ([], MCPSessionManager(), [])

        project_context = ProjectContext.from_user_cwd(user_cwd)
        await resolve_and_load_mcp_tools(
            explicit_config_path="configs/mcp.json",
            trust_project_mcp=True,
            project_context=project_context,
        )

        merged = mock_load.call_args.args[0]
        assert "fs" in merged["mcpServers"]

    @patch("deepagents_cli.mcp_tools._load_tools_from_config")
    @patch("deepagents_cli.mcp_tools.discover_mcp_configs")
    async def test_trust_false_filters_project_stdio(
        self,
        mock_discover: MagicMock,
        mock_load: AsyncMock,
        tmp_path: Path,
    ) -> None:
        """trust_project_mcp=False filters out project-level stdio servers."""
        project_cfg = tmp_path / ".mcp.json"
        project_cfg.write_text(
            json.dumps(
                {
                    "mcpServers": {
                        "pwn": {"command": "bash", "args": ["-c", "evil"]},
                        "remote": {"type": "sse", "url": "http://ok"},
                    }
                }
            )
        )
        mock_discover.return_value = [project_cfg]
        mock_load.return_value = ([], MCPSessionManager(), [])

        await resolve_and_load_mcp_tools(trust_project_mcp=False)

        mock_load.assert_awaited_once()
        merged = mock_load.call_args.args[0]
        assert "pwn" not in merged["mcpServers"]
        assert "remote" in merged["mcpServers"]

    @patch("deepagents_cli.mcp_tools._load_tools_from_config")
    @patch("deepagents_cli.mcp_tools.discover_mcp_configs")
    async def test_trust_true_allows_project_stdio(
        self,
        mock_discover: MagicMock,
        mock_load: AsyncMock,
        tmp_path: Path,
    ) -> None:
        """trust_project_mcp=True allows project-level stdio servers."""
        project_cfg = tmp_path / ".mcp.json"
        project_cfg.write_text(
            json.dumps({"mcpServers": {"local": {"command": "npx", "args": []}}})
        )
        mock_discover.return_value = [project_cfg]
        mock_load.return_value = ([], MCPSessionManager(), [])

        await resolve_and_load_mcp_tools(trust_project_mcp=True)

        mock_load.assert_awaited_once()
        merged = mock_load.call_args.args[0]
        assert "local" in merged["mcpServers"]


class TestClassifyDiscoveredConfigs:
    """Tests for classify_discovered_configs."""

    def test_user_config_classified(self) -> None:
        """Paths under ~/.deepagents/ are classified as user."""
        user_path = Path.home() / ".deepagents" / ".mcp.json"
        user, project = classify_discovered_configs([user_path])
        assert user == [user_path]
        assert project == []

    def test_project_config_classified(self, tmp_path: Path) -> None:
        """Paths outside ~/.deepagents/ are classified as project."""
        project_path = tmp_path / ".mcp.json"
        project_path.touch()
        user, project = classify_discovered_configs([project_path])
        assert user == []
        assert project == [project_path]

    def test_mixed_classification(self, tmp_path: Path) -> None:
        """Mixed paths are split correctly."""
        user_path = Path.home() / ".deepagents" / ".mcp.json"
        project_path = tmp_path / ".mcp.json"
        project_path.touch()
        user, project = classify_discovered_configs([user_path, project_path])
        assert user == [user_path]
        assert project == [project_path]


class TestExtractStdioServerCommands:
    """Tests for extract_stdio_server_commands."""

    def test_extracts_stdio(self) -> None:
        """Extracts name, command, args from stdio servers."""
        config = {
            "mcpServers": {
                "fs": {"command": "npx", "args": ["-y", "fs-server"]},
            }
        }
        result = extract_stdio_server_commands(config)
        assert result == [("fs", "npx", ["-y", "fs-server"])]

    def test_skips_remote(self) -> None:
        """SSE/HTTP servers are not extracted."""
        config = {
            "mcpServers": {
                "remote": {"type": "sse", "url": "http://example.com"},
            }
        }
        assert extract_stdio_server_commands(config) == []

    def test_mixed(self) -> None:
        """Only stdio servers are returned from mixed configs."""
        config = {
            "mcpServers": {
                "local": {"command": "bash", "args": []},
                "remote": {"type": "http", "url": "http://x"},
            }
        }
        result = extract_stdio_server_commands(config)
        assert len(result) == 1
        assert result[0][0] == "local"

    def test_empty_servers(self) -> None:
        """Empty mcpServers returns empty list."""
        assert extract_stdio_server_commands({"mcpServers": {}}) == []

    def test_no_servers_key(self) -> None:
        """Missing mcpServers returns empty list."""
        assert extract_stdio_server_commands({}) == []


class TestFilterProjectStdioServers:
    """Tests for _filter_project_stdio_servers."""

    def test_removes_stdio_keeps_remote(self) -> None:
        """Stdio servers are removed, remote servers are kept."""
        config = {
            "mcpServers": {
                "local": {"command": "bash", "args": []},
                "remote": {"type": "sse", "url": "http://x"},
            }
        }
        result = _filter_project_stdio_servers(config)
        assert "local" not in result["mcpServers"]
        assert "remote" in result["mcpServers"]

    def test_all_stdio_returns_empty(self) -> None:
        """Config with only stdio servers returns empty mcpServers."""
        config = {"mcpServers": {"a": {"command": "x", "args": []}}}
        result = _filter_project_stdio_servers(config)
        assert result["mcpServers"] == {}


class TestCheckStdioServer:
    """Tests for _check_stdio_server pre-flight validation."""

    def test_command_not_found(self) -> None:
        """Raises RuntimeError with server name and command when missing."""
        with (
            patch("deepagents_cli.mcp_tools.shutil.which", return_value=None),
            pytest.raises(
                RuntimeError,
                match="MCP server 'test-server': command 'nonexistent' not found",
            ),
        ):
            _check_stdio_server("test-server", {"command": "nonexistent"})

    def test_command_exists(self) -> None:
        """No error when command exists on PATH."""
        with patch(
            "deepagents_cli.mcp_tools.shutil.which", return_value="/usr/bin/npx"
        ):
            _check_stdio_server("test-server", {"command": "npx"})

    def test_missing_command_key(self) -> None:
        """Raises RuntimeError when config lacks `command` key."""
        with pytest.raises(RuntimeError, match="missing 'command' in config"):
            _check_stdio_server("test-server", {})


class TestCheckRemoteServer:
    """Tests for _check_remote_server pre-flight validation."""

    async def test_unreachable(self) -> None:
        """Raises RuntimeError when URL is unreachable."""
        import httpx

        mock_client = AsyncMock()
        mock_client.head.side_effect = httpx.TransportError("connection refused")
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock(return_value=False)

        with (
            patch("httpx.AsyncClient", return_value=mock_client),
            pytest.raises(RuntimeError, match="unreachable"),
        ):
            await _check_remote_server("test-server", {"url": "http://localhost:9999"})

    async def test_reachable(self) -> None:
        """No error when URL responds."""
        mock_client = AsyncMock()
        mock_client.head = AsyncMock()
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock(return_value=False)

        with patch("httpx.AsyncClient", return_value=mock_client):
            await _check_remote_server("test-server", {"url": "http://localhost:8080"})

        mock_client.head.assert_awaited_once_with("http://localhost:8080", timeout=2)

    async def test_unreachable_os_error(self) -> None:
        """Raises RuntimeError on OS-level socket errors."""
        mock_client = AsyncMock()
        mock_client.head.side_effect = OSError("Network is unreachable")
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock(return_value=False)

        with (
            patch("httpx.AsyncClient", return_value=mock_client),
            pytest.raises(RuntimeError, match="unreachable"),
        ):
            await _check_remote_server(
                "test-server", {"url": "http://10.255.255.1:9999"}
            )

    async def test_invalid_url(self) -> None:
        """Raises RuntimeError on malformed URLs."""
        import httpx

        mock_client = AsyncMock()
        mock_client.head.side_effect = httpx.InvalidURL("Invalid URL")
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock(return_value=False)

        with (
            patch("httpx.AsyncClient", return_value=mock_client),
            pytest.raises(RuntimeError, match="unreachable"),
        ):
            await _check_remote_server("test-server", {"url": "http://"})

    async def test_missing_url_key(self) -> None:
        """Raises RuntimeError when config lacks `url` key."""
        with pytest.raises(RuntimeError, match="missing 'url' in config"):
            await _check_remote_server("test-server", {})

    async def test_timeout(self) -> None:
        """Raises RuntimeError on connection timeout."""
        import httpx

        mock_client = AsyncMock()
        mock_client.head.side_effect = httpx.TimeoutException("timed out")
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock(return_value=False)

        with (
            patch("httpx.AsyncClient", return_value=mock_client),
            pytest.raises(RuntimeError, match="unreachable"),
        ):
            await _check_remote_server("test-server", {"url": "http://slow:8080"})


class TestHealthCheckIntegration:
    """Tests for health check integration in _load_tools_from_config."""

    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_stdio_health_check_failure_skips_session(
        self,
        mock_client_class: MagicMock,
        write_config: Callable[..., str],
    ) -> None:
        """Health check failure cleans up and does not call client.session()."""
        path = write_config(
            {"mcpServers": {"fs": {"command": "missing-cmd", "args": []}}}
        )
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        with (
            patch("deepagents_cli.mcp_tools.shutil.which", return_value=None),
            pytest.raises(RuntimeError, match="Pre-flight health check"),
        ):
            await get_mcp_tools(path)

        # session() should never be called — health check fails first
        mock_client.session.assert_not_called()

    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_remote_health_check_failure_skips_session(
        self,
        mock_client_class: MagicMock,
        write_config: Callable[..., str],
    ) -> None:
        """Remote health check failure prevents session creation."""
        import httpx

        path = write_config(
            {"mcpServers": {"api": {"type": "sse", "url": "http://down:9999"}}}
        )
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        mock_http = AsyncMock()
        mock_http.head.side_effect = httpx.TransportError("refused")
        mock_http.__aenter__ = AsyncMock(return_value=mock_http)
        mock_http.__aexit__ = AsyncMock(return_value=False)

        with (
            patch("httpx.AsyncClient", return_value=mock_http),
            pytest.raises(RuntimeError, match="Pre-flight health check"),
        ):
            await get_mcp_tools(path)

        mock_client.session.assert_not_called()

    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_multi_server_collects_all_failures(
        self,
        mock_client_class: MagicMock,
        write_config: Callable[..., str],
    ) -> None:
        """All server failures are reported in a single error."""
        path = write_config(
            {
                "mcpServers": {
                    "a": {"command": "missing-a", "args": []},
                    "b": {"command": "missing-b", "args": []},
                }
            }
        )
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        with (
            patch("deepagents_cli.mcp_tools.shutil.which", return_value=None),
            pytest.raises(RuntimeError) as exc_info,
        ):
            await get_mcp_tools(path)

        error_msg = str(exc_info.value)
        assert "missing-a" in error_msg
        assert "missing-b" in error_msg
        mock_client.session.assert_not_called()

    @patch("langchain_mcp_adapters.client.MultiServerMCPClient")
    async def test_mixed_stdio_and_remote_checks(
        self,
        mock_client_class: MagicMock,
        write_config: Callable[..., str],
    ) -> None:
        """Both stdio and remote health checks run for mixed configs."""
        import httpx

        path = write_config(
            {
                "mcpServers": {
                    "local": {"command": "missing-cmd", "args": []},
                    "remote": {"type": "sse", "url": "http://down:9999"},
                }
            }
        )
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        mock_http = AsyncMock()
        mock_http.head.side_effect = httpx.TransportError("refused")
        mock_http.__aenter__ = AsyncMock(return_value=mock_http)
        mock_http.__aexit__ = AsyncMock(return_value=False)

        with (
            patch("deepagents_cli.mcp_tools.shutil.which", return_value=None),
            patch("httpx.AsyncClient", return_value=mock_http),
            pytest.raises(RuntimeError) as exc_info,
        ):
            await get_mcp_tools(path)

        # Both failures appear in the error message
        error_msg = str(exc_info.value)
        assert "missing-cmd" in error_msg
        assert "down:9999" in error_msg



"""Tests for server graph MCP loading behavior."""

from __future__ import annotations

import importlib
import os
import sys
from types import ModuleType, SimpleNamespace
from unittest.mock import AsyncMock, MagicMock, patch

from deepagents_cli._server_config import ServerConfig
from deepagents_cli._server_constants import ENV_PREFIX


def _import_fresh_server_graph() -> ModuleType:
    """Import `deepagents_cli.server_graph` from a clean module state."""
    sys.modules.pop("deepagents_cli.server_graph", None)
    return importlib.import_module("deepagents_cli.server_graph")


def _module_with_attrs(name: str, **attrs: object) -> ModuleType:
    """Create a module stub with dynamically assigned attributes."""
    module = ModuleType(name)
    for key, value in attrs.items():
        setattr(module, key, value)
    return module


class TestServerGraph:
    """Tests for server-mode graph bootstrap."""

    def test_auto_discovery_loads_mcp_without_explicit_config(self) -> None:
        """Server mode should auto-discover MCP configs when no path is passed."""
        graph_obj = object()
        model_obj = object()
        http_tool = object()
        fetch_tool = object()
        mcp_tool = object()
        mcp_server_info = [SimpleNamespace(name="docs")]
        create_cli_agent = MagicMock(return_value=(graph_obj, object()))
        agent_module = _module_with_attrs(
            "deepagents_cli.agent",
            DEFAULT_AGENT_NAME="agent",
            create_cli_agent=create_cli_agent,
            load_async_subagents=MagicMock(return_value=None),
        )

        model_result = SimpleNamespace(
            model=model_obj,
            apply_to_settings=MagicMock(),
        )
        config_module = _module_with_attrs(
            "deepagents_cli.config",
            create_model=MagicMock(return_value=model_result),
            settings=SimpleNamespace(
                has_tavily=False,
                reload_from_environment=MagicMock(),
            ),
        )

        tools_module = _module_with_attrs(
            "deepagents_cli.tools",
            http_request=http_tool,
            fetch_url=fetch_tool,
            web_search=object(),
        )

        resolve_mcp_tools = AsyncMock(return_value=([mcp_tool], None, mcp_server_info))
        mcp_module = _module_with_attrs(
            "deepagents_cli.mcp_tools",
            resolve_and_load_mcp_tools=resolve_mcp_tools,
        )

        # Build env from ServerConfig to exercise the same serialization
        # path the real CLI uses.
        config = ServerConfig(no_mcp=False)
        env_overrides = {}
        for suffix, value in config.to_env().items():
            if value is not None:
                env_overrides[f"{ENV_PREFIX}{suffix}"] = value

        with (
            patch.dict(os.environ, env_overrides, clear=False),
            patch.dict(
                sys.modules,
                {
                    "deepagents_cli.agent": agent_module,
                    "deepagents_cli.config": config_module,
                    "deepagents_cli.tools": tools_module,
                    "deepagents_cli.mcp_tools": mcp_module,
                },
            ),
            patch(
                "deepagents_cli.project_utils.get_server_project_context",
                return_value=None,
            ),
        ):
            for suffix in (
                "MCP_CONFIG_PATH",
                "TRUST_PROJECT_MCP",
                "CWD",
                "PROJECT_ROOT",
            ):
                os.environ.pop(f"{ENV_PREFIX}{suffix}", None)

            module = _import_fresh_server_graph()

        resolve_mcp_tools.assert_awaited_once_with(
            explicit_config_path=None,
            no_mcp=False,
            trust_project_mcp=None,
            project_context=None,
        )
        create_cli_agent.assert_called_once_with(
            model=model_obj,
            assistant_id="agent",
            tools=[http_tool, fetch_tool, mcp_tool],
            sandbox=None,
            sandbox_type=None,
            system_prompt=None,
            interactive=True,
            auto_approve=False,
            enable_ask_user=False,
            enable_memory=True,
            enable_skills=True,
            enable_shell=True,
            mcp_server_info=mcp_server_info,
            cwd=None,
            project_context=None,
            async_subagents=None,
        )
        assert module.graph is graph_obj



"""Tests for shell command allow-list functionality."""

import pytest

from deepagents_cli.config import (
    SHELL_ALLOW_ALL,
    contains_dangerous_patterns,
    is_shell_command_allowed,
)


@pytest.fixture
def basic_allow_list() -> list[str]:
    """Basic allow-list with common read-only commands."""
    return ["ls", "cat", "grep"]


@pytest.fixture
def extended_allow_list() -> list[str]:
    """Extended allow-list with common read-only commands."""
    return ["ls", "cat", "grep", "wc", "pwd", "echo", "head", "tail", "find", "sort"]


@pytest.fixture
def semicolon_allow_list() -> list[str]:
    """Allow-list for semicolon-separated command tests."""
    return ["ls", "cat", "pwd"]


@pytest.fixture
def quoted_allow_list() -> list[str]:
    """Allow-list for quoted argument tests."""
    return ["echo", "grep"]


@pytest.fixture
def pipeline_allow_list() -> list[str]:
    """Allow-list for complex pipeline tests."""
    return ["cat", "grep", "wc", "sort"]


@pytest.mark.parametrize(
    "command",
    ["ls", "cat file.txt", "rm -rf /"],
)
def test_empty_allow_list_rejects_all(command: str) -> None:
    """Test that empty allow-list rejects all commands."""
    assert not is_shell_command_allowed(command, [])


@pytest.mark.parametrize(
    "command",
    ["ls", "cat file.txt"],
)
def test_none_allow_list_rejects_all(command: str) -> None:
    """Test that None allow-list rejects all commands."""
    assert not is_shell_command_allowed(command, None)


@pytest.mark.parametrize(
    "command",
    ["ls", "cat", "grep"],
)
def test_simple_command_allowed(command: str, basic_allow_list: list[str]) -> None:
    """Test simple commands that are in the allow-list."""
    assert is_shell_command_allowed(command, basic_allow_list)


@pytest.mark.parametrize(
    "command",
    ["rm", "mv", "chmod"],
)
def test_simple_command_not_allowed(command: str, basic_allow_list: list[str]) -> None:
    """Test simple commands that are not in the allow-list."""
    assert not is_shell_command_allowed(command, basic_allow_list)


@pytest.mark.parametrize(
    "command",
    [
        "ls -la",
        "cat file.txt",
        "grep 'pattern' file.txt",
        "ls -la /tmp/test",
    ],
)
def test_command_with_arguments_allowed(
    command: str, basic_allow_list: list[str]
) -> None:
    """Test commands with arguments that are in the allow-list."""
    assert is_shell_command_allowed(command, basic_allow_list)


@pytest.mark.parametrize(
    "command",
    ["rm -rf /tmp", "mv file1 file2"],
)
def test_command_with_arguments_not_allowed(
    command: str, basic_allow_list: list[str]
) -> None:
    """Test commands with arguments that are not in the allow-list."""
    assert not is_shell_command_allowed(command, basic_allow_list)


@pytest.mark.parametrize(
    ("command", "allow_list"),
    [
        ("ls | grep test", ["ls", "cat", "grep", "wc"]),
        ("cat file.txt | grep pattern", ["ls", "cat", "grep", "wc"]),
        ("ls -la | wc -l", ["ls", "cat", "grep", "wc"]),
        ("cat file | grep foo | wc", ["ls", "cat", "grep", "wc"]),
    ],
)
def test_piped_commands_all_allowed(command: str, allow_list: list[str]) -> None:
    """Test piped commands where all parts are in the allow-list."""
    assert is_shell_command_allowed(command, allow_list)


@pytest.mark.parametrize(
    "command",
    [
        "ls | wc -l",  # wc not in allow-list
        "cat file.txt | sort",  # sort not in allow-list
        "grep pattern file | rm",  # rm not in allow-list
    ],
)
def test_piped_commands_some_not_allowed(
    command: str, basic_allow_list: list[str]
) -> None:
    """Test piped commands where some parts are not in the allow-list."""
    assert not is_shell_command_allowed(command, basic_allow_list)


@pytest.mark.parametrize(
    "command",
    ["ls; pwd", "pwd; ls -la"],
)
def test_semicolon_separated_commands_all_allowed(
    command: str, semicolon_allow_list: list[str]
) -> None:
    """Test semicolon-separated commands where all are in the allow-list."""
    assert is_shell_command_allowed(command, semicolon_allow_list)


@pytest.mark.parametrize(
    "command",
    ["ls; rm file", "pwd; mv file1 file2"],
)
def test_semicolon_separated_commands_some_not_allowed(
    command: str, semicolon_allow_list: list[str]
) -> None:
    """Test semicolon-separated commands where some are not in the allow-list."""
    assert not is_shell_command_allowed(command, semicolon_allow_list)


@pytest.mark.parametrize(
    ("command", "expected"),
    [
        ("ls && cat file", True),
        ("ls && rm file", False),
        ("ls -la && grep pattern file && cat output.txt", True),
        ("ls && cat file && grep test", True),
        ("cat a.txt && cat b.txt && cat c.txt", True),
        ("ls && rm -rf /", False),
    ],
)
def test_and_operator_commands(
    command: str, *, expected: bool, basic_allow_list: list[str]
) -> None:
    """Test commands with && operator (commonly used by Claude for chaining)."""
    assert is_shell_command_allowed(command, basic_allow_list) == expected


@pytest.mark.parametrize(
    ("command", "expected"),
    [
        ("ls || cat file", True),
        ("ls || rm file", False),
    ],
)
def test_or_operator_commands(
    command: str, *, expected: bool, basic_allow_list: list[str]
) -> None:
    """Test commands with || operator."""
    assert is_shell_command_allowed(command, basic_allow_list) == expected


@pytest.mark.parametrize(
    "command",
    [
        'echo "hello world"',
        "grep 'pattern' file.txt",
        'echo "test" | grep "te"',
    ],
)
def test_quoted_arguments(command: str, quoted_allow_list: list[str]) -> None:
    """Test commands with quoted arguments."""
    assert is_shell_command_allowed(command, quoted_allow_list)


@pytest.mark.parametrize(
    "command",
    ["  ls  ", "ls   -la", "ls | cat"],
)
def test_whitespace_handling(command: str, basic_allow_list: list[str]) -> None:
    """Test proper handling of whitespace."""
    assert is_shell_command_allowed(command, basic_allow_list)


@pytest.mark.parametrize(
    "command",
    ['ls "unclosed', "cat 'missing quote"],
)
def test_malformed_commands_rejected(command: str, basic_allow_list: list[str]) -> None:
    """Test that malformed commands are rejected for safety."""
    assert not is_shell_command_allowed(command, basic_allow_list)


@pytest.mark.parametrize(
    "command",
    ["", "   "],
)
def test_empty_command_rejected(command: str, basic_allow_list: list[str]) -> None:
    """Test that empty commands are rejected."""
    assert not is_shell_command_allowed(command, basic_allow_list)


@pytest.mark.parametrize(
    ("command", "expected"),
    [
        ("ls", True),
        ("LS", False),
        ("Cat", False),
    ],
)
def test_case_sensitivity(
    command: str, *, expected: bool, basic_allow_list: list[str]
) -> None:
    """Test that command matching is case-sensitive."""
    assert is_shell_command_allowed(command, basic_allow_list) == expected


@pytest.mark.parametrize(
    "command",
    [
        "ls -la",
        "cat file.txt",
        "grep pattern file",
        "pwd",
        "echo 'hello'",
        "head -n 10 file",
        "tail -f log.txt",
        "find . -name '*.py'",
        "wc -l file",
    ],
)
def test_common_read_only_commands(
    command: str, extended_allow_list: list[str]
) -> None:
    """Test common read-only commands are allowed."""
    assert is_shell_command_allowed(command, extended_allow_list)


@pytest.mark.parametrize(
    "command",
    [
        "rm -rf /",
        "mv file /dev/null",
        "chmod 777 file",
        "dd if=/dev/zero of=/dev/sda",
        "mkfs.ext4 /dev/sda",
    ],
)
def test_dangerous_commands_not_allowed(
    command: str, basic_allow_list: list[str]
) -> None:
    """Test that dangerous commands are not allowed by default."""
    assert not is_shell_command_allowed(command, basic_allow_list)


@pytest.mark.parametrize(
    ("command", "expected"),
    [
        ("cat file.txt | grep error | sort | wc -l", True),
        ("cat file.txt | grep error | rm | wc -l", False),
    ],
)
def test_complex_pipeline(
    command: str, *, expected: bool, pipeline_allow_list: list[str]
) -> None:
    """Test complex pipelines with multiple operators."""
    assert is_shell_command_allowed(command, pipeline_allow_list) == expected


class TestShellInjectionPrevention:
    """Tests for shell injection attack prevention."""

    @pytest.fixture
    def injection_allow_list(self) -> list[str]:
        """Allow-list for injection tests."""
        return ["ls", "cat", "grep", "echo"]

    @pytest.mark.parametrize(
        "command",
        [
            "ls $(rm -rf /)",
            "ls $(cat /etc/shadow)",
            'echo "$(whoami)"',
            "cat $(echo /etc/passwd)",
        ],
    )
    def test_command_substitution_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Command substitution $(...) must be blocked even in arguments."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "ls `rm -rf /`",
            "ls `cat /etc/shadow`",
            "echo `whoami`",
            "cat `echo /etc/passwd`",
        ],
    )
    def test_backtick_substitution_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Backtick command substitution must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "ls\nrm -rf /",
            "cat file\nwhoami",
            "echo hello\n/bin/sh",
        ],
    )
    def test_newline_injection_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Newline characters must be blocked (command injection)."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "ls\rrm -rf /",
            "cat file\rwhoami",
        ],
    )
    def test_carriage_return_injection_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Carriage return characters must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "ls\trm",
            "cat\t/etc/passwd",
        ],
    )
    def test_tab_injection_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Tab characters must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "cat <(rm -rf /)",
            "cat <(whoami)",
            "grep pattern <(cat /etc/shadow)",
        ],
    )
    def test_process_substitution_input_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Process substitution <(...) must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "cat >(rm -rf /)",
            "echo test >(cat)",
        ],
    )
    def test_process_substitution_output_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Process substitution >(...) must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "cat <<< 'rm -rf /'",
            "grep pattern <<< 'test'",
        ],
    )
    def test_here_string_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Here-strings (<<<) must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "cat << EOF\nmalicious\nEOF",
            "cat <<EOF",
        ],
    )
    def test_here_doc_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Here-documents (<<) must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "echo hello > /tmp/test",
            "ls > output.txt",
            "cat file > /etc/passwd",
        ],
    )
    def test_output_redirect_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Output redirection (>) must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "echo hello >> /tmp/test",
            "ls >> output.txt",
        ],
    )
    def test_append_redirect_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Append redirection (>>) must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "cat < /etc/passwd",
            "grep pattern < input.txt",
        ],
    )
    def test_input_redirect_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Input redirection (<) must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "echo ${PATH}",
            "cat ${HOME}/.bashrc",
            "ls ${PWD}",
        ],
    )
    def test_brace_variable_expansion_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Brace variable expansion ${...} must be blocked (can contain commands)."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "ls -la",
            "cat file.txt",
            "grep pattern file",
            "echo hello world",
            "ls | grep test",
            "cat file | grep pattern",
        ],
    )
    def test_safe_commands_still_allowed(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Verify that safe commands without dangerous patterns are still allowed."""
        assert is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "/bin/rm -rf /",
            "./malicious",
            "../../../bin/sh",
        ],
    )
    def test_path_bypass_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """Commands with paths (not in allow-list) must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            r"cat $'\050whoami\051'",
            r"cat $'\044\050whoami\051'",
            r"echo $'\140whoami\140'",
            r"cat $'\x24\x28whoami\x29'",
            r"echo $'\076/tmp/evil'",
            r"cat $'\074/etc/passwd'",
        ],
    )
    def test_ansi_c_quoting_blocked(
        self, command: str, injection_allow_list: list[str]
    ) -> None:
        """ANSI-C quoting $'...' with escape sequences must be blocked."""
        assert not is_shell_command_allowed(command, injection_allow_list)


class TestBackgroundOperatorBlocking:
    """Tests for standalone & (background operator) detection."""

    @pytest.mark.parametrize(
        "command",
        [
            "ls &",
            "cat file.txt & echo done",
            "sleep 10 &",
        ],
    )
    def test_background_operator_blocked(
        self, command: str, basic_allow_list: list[str]
    ) -> None:
        """Standalone & (background execution) must be blocked."""
        assert not is_shell_command_allowed(command, basic_allow_list)

    @pytest.mark.parametrize(
        ("command", "expected"),
        [
            ("ls && cat file", True),
            ("ls && grep test file", True),
        ],
    )
    def test_double_ampersand_still_works(
        self, command: str, *, expected: bool, basic_allow_list: list[str]
    ) -> None:
        """Double && (AND operator) should still be allowed."""
        assert is_shell_command_allowed(command, basic_allow_list) == expected


class TestBareVariableBlocking:
    """Tests for bare $VARIABLE expansion detection."""

    @pytest.mark.parametrize(
        "command",
        [
            "ls $HOME",
            "cat $PATH",
            "echo $USER",
            "grep $IFS file.txt",
        ],
    )
    def test_bare_variable_blocked(
        self, command: str, basic_allow_list: list[str]
    ) -> None:
        """Bare $VARIABLE expansion must be blocked to prevent info leaks."""
        assert not is_shell_command_allowed(command, basic_allow_list)

    @pytest.mark.parametrize(
        "command",
        [
            "ls -la",
            "cat file.txt",
            "grep pattern file",
        ],
    )
    def test_commands_without_variables_still_work(
        self, command: str, basic_allow_list: list[str]
    ) -> None:
        """Commands without variable expansion should still be allowed."""
        assert is_shell_command_allowed(command, basic_allow_list)


class TestContainsDangerousPatterns:
    """Direct tests for contains_dangerous_patterns()."""

    @pytest.mark.parametrize(
        "command",
        [
            "ls -la",
            "cat file.txt",
            "grep pattern file",
            "wc -l file.txt",
            "head -n 10 file.txt",
            "echo hello world",
        ],
    )
    def test_safe_commands_pass(self, command: str) -> None:
        """Safe commands should not be flagged as dangerous."""
        assert not contains_dangerous_patterns(command)

    @pytest.mark.parametrize(
        ("command", "description"),
        [
            ("ls $(whoami)", "command substitution with $()"),
            ("cat `whoami`", "backtick command substitution"),
            ("echo $'\\x41'", "ANSI-C quoting"),
            ("ls\nrm -rf /", "newline injection"),
            ("ls\rrm -rf /", "carriage return injection"),
            ("ls\trm", "tab injection"),
            ("cat <(ls)", "process substitution (input)"),
            ("cat >(ls)", "process substitution (output)"),
            ("cat <<EOF", "here-doc"),
            ("cat <<<word", "here-string"),
            ("ls >> file", "append redirect"),
            ("ls > file", "output redirect"),
            ("cat < file", "input redirect"),
            ("echo ${PATH}", "variable expansion with braces"),
            ("ls $HOME", "bare variable expansion"),
            ("ls &", "background operator"),
            ("sleep 10 &", "trailing background operator"),
        ],
    )
    def test_dangerous_patterns_detected(self, command: str, description: str) -> None:
        """Dangerous shell patterns should be detected."""
        assert contains_dangerous_patterns(command), f"Failed to detect: {description}"

    def test_double_ampersand_not_flagged(self) -> None:
        """Double && should not be flagged as dangerous (it's a safe operator)."""
        assert not contains_dangerous_patterns("ls && cat file")

    def test_empty_command(self) -> None:
        """Empty command should not be flagged."""
        assert not contains_dangerous_patterns("")

    def test_pattern_in_middle_of_command(self) -> None:
        """Dangerous patterns embedded within arguments should be detected."""
        assert contains_dangerous_patterns("echo foo$(bar)baz")

    def test_multiple_dangerous_patterns(self) -> None:
        """Commands with multiple dangerous patterns should be detected."""
        assert contains_dangerous_patterns("cat $(cmd) > /tmp/out")


class TestFindExecLimitation:
    """Document that `find -exec` is NOT caught by `contains_dangerous_patterns`.

    `find -exec` enables arbitrary command execution but uses a flag rather than
    a shell metacharacter, so the substring-based pattern check cannot detect it.
    This is why `find` is excluded from `RECOMMENDED_SAFE_SHELL_COMMANDS`. Users
    who add `find` to a custom allow-list accept this risk.
    """

    def test_find_exec_not_caught_by_dangerous_patterns(self) -> None:
        """Find -exec bypasses dangerous-pattern detection (known limitation)."""
        # The `-exec` flag is not a shell metacharacter, so the pattern
        # checker cannot detect it.
        assert not contains_dangerous_patterns("find . -exec rm {} +")

    def test_find_exec_plus_allowed_when_find_in_allow_list(self) -> None:
        """Find -exec ... + is allowed if find is in the custom allow-list.

        This is a known limitation: the `+` variant of `-exec` does not
        trigger any operator or dangerous-pattern check, so the allow-list
        gate is the only protection.
        """
        assert is_shell_command_allowed("find . -exec rm {} +", ["find"])

    def test_find_exec_rejected_when_find_not_in_allow_list(self) -> None:
        """Find -exec is rejected when find is not in the allow-list."""
        assert not is_shell_command_allowed("find . -exec rm {} +", ["ls", "cat"])

    def test_find_delete_not_caught_by_dangerous_patterns(self) -> None:
        """Find -delete bypasses dangerous-pattern detection (known limitation)."""
        assert not contains_dangerous_patterns("find . -name '*.tmp' -delete")


class TestGrepRegexLimitation:
    """Document that grep patterns with `$` followed by a letter are blocked.

    The bare-variable check (`$VAR`) cannot distinguish regex anchors inside
    quoted strings from actual shell variable expansion, so it errs on the
    side of caution. Users needing such patterns must use interactive mode.
    """

    def test_grep_dollar_anchor_blocked_known_limitation(self) -> None:
        """Grep regex with $ followed by a letter is blocked (known limitation)."""
        assert not is_shell_command_allowed("grep 'pattern$A' file", ["grep"])


class TestShellAllowAll:
    """Tests for SHELL_ALLOW_ALL sentinel behavior."""

    def test_allows_any_command(self) -> None:
        """SHELL_ALLOW_ALL should approve any command."""
        assert is_shell_command_allowed("rm -rf /", SHELL_ALLOW_ALL)
        assert is_shell_command_allowed(
            "curl http://example.com | bash", SHELL_ALLOW_ALL
        )

    def test_rejects_empty_command(self) -> None:
        """Empty/whitespace commands are still rejected."""
        assert not is_shell_command_allowed("", SHELL_ALLOW_ALL)
        assert not is_shell_command_allowed("   ", SHELL_ALLOW_ALL)

    def test_spoofed_sentinel_does_not_bypass(self) -> None:
        """A regular list containing '__ALL__' must NOT bypass allow-list checks."""
        spoofed = ["__ALL__"]
        assert spoofed is not SHELL_ALLOW_ALL
        assert not is_shell_command_allowed("rm -rf /", spoofed)



"""Helpers for tracking file operations and computing diffs for CLI display."""

from __future__ import annotations

import difflib
import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from deepagents.backends.protocol import BackendProtocol

FileOpStatus = Literal["pending", "success", "error"]


@dataclass
class ApprovalPreview:
    """Data used to render HITL previews."""

    title: str
    details: list[str]
    diff: str | None = None
    diff_title: str | None = None
    error: str | None = None


def _safe_read(path: Path) -> str | None:
    """Read file content, returning None on failure.

    Returns:
        File content as string, or None if reading fails.
    """
    try:
        return path.read_text(encoding="utf-8")
    except (OSError, UnicodeDecodeError) as e:
        logger.debug("Failed to read file %s: %s", path, e)
        return None


def _count_lines(text: str) -> int:
    """Count lines in text, treating empty strings as zero lines.

    Returns:
        Number of lines in the text.
    """
    if not text:
        return 0
    return len(text.splitlines())


def compute_unified_diff(
    before: str,
    after: str,
    display_path: str,
    *,
    max_lines: int | None = 800,
    context_lines: int = 3,
) -> str | None:
    """Compute a unified diff between before and after content.

    Args:
        before: Original content
        after: New content
        display_path: Path for display in diff headers
        max_lines: Maximum number of diff lines (None for unlimited)
        context_lines: Number of context lines around changes (default 3)

    Returns:
        Unified diff string or None if no changes
    """
    before_lines = before.splitlines()
    after_lines = after.splitlines()
    diff_lines = list(
        difflib.unified_diff(
            before_lines,
            after_lines,
            fromfile=f"{display_path} (before)",
            tofile=f"{display_path} (after)",
            lineterm="",
            n=context_lines,
        )
    )
    if not diff_lines:
        return None
    if max_lines is not None and len(diff_lines) > max_lines:
        truncated = diff_lines[: max_lines - 1]
        truncated.append("...")
        return "\n".join(truncated)
    return "\n".join(diff_lines)


@dataclass
class FileOpMetrics:
    """Line and byte level metrics for a file operation."""

    lines_read: int = 0
    start_line: int | None = None
    end_line: int | None = None
    lines_written: int = 0
    lines_added: int = 0
    lines_removed: int = 0
    bytes_written: int = 0


@dataclass
class FileOperationRecord:
    """Track a single filesystem tool call."""

    tool_name: str
    display_path: str
    physical_path: Path | None
    tool_call_id: str | None
    args: dict[str, Any] = field(default_factory=dict)
    status: FileOpStatus = "pending"
    error: str | None = None
    metrics: FileOpMetrics = field(default_factory=FileOpMetrics)
    diff: str | None = None
    before_content: str | None = None
    after_content: str | None = None
    read_output: str | None = None
    hitl_approved: bool = False


def resolve_physical_path(
    path_str: str | None, assistant_id: str | None
) -> Path | None:
    """Convert a virtual/relative path to a physical filesystem path.

    Returns:
        Resolved physical Path, or None if path is empty or resolution fails.
    """
    if not path_str:
        return None
    try:
        if assistant_id and path_str.startswith("/memories/"):
            from deepagents_cli.config import settings

            agent_dir = settings.get_agent_dir(assistant_id)
            suffix = path_str.removeprefix("/memories/").lstrip("/")
            return (agent_dir / suffix).resolve()
        path = Path(path_str)
        if path.is_absolute():
            return path
        return (Path.cwd() / path).resolve()
    except (OSError, ValueError):
        return None


def format_display_path(path_str: str | None) -> str:
    """Format a path for display.

    Returns:
        Formatted path string suitable for display.
    """
    if not path_str:
        return "(unknown)"
    try:
        path = Path(path_str)
        if path.is_absolute():
            return path.name or str(path)
        return str(path)
    except (OSError, ValueError):
        return str(path_str)


def build_approval_preview(
    tool_name: str,
    args: dict[str, Any],
    assistant_id: str | None,
) -> ApprovalPreview | None:
    """Collect summary info and diff for HITL approvals.

    Returns:
        ApprovalPreview with diff and details, or None if tool not supported.
    """
    path_str = str(args.get("file_path") or args.get("path") or "")
    display_path = format_display_path(path_str)
    physical_path = resolve_physical_path(path_str, assistant_id)

    if tool_name == "write_file":
        content = str(args.get("content", ""))
        before = (
            _safe_read(physical_path)
            if physical_path and physical_path.exists()
            else ""
        )
        after = content
        diff = compute_unified_diff(before or "", after, display_path, max_lines=100)
        additions = 0
        if diff:
            additions = sum(
                1
                for line in diff.splitlines()
                if line.startswith("+") and not line.startswith("+++")
            )
        total_lines = _count_lines(after)
        details = [
            f"File: {path_str}",
            "Action: Create new file"
            + (" (overwrites existing content)" if before else ""),
            f"Lines to write: {additions or total_lines}",
        ]
        return ApprovalPreview(
            title=f"Write {display_path}",
            details=details,
            diff=diff,
            diff_title=f"Diff {display_path}",
        )

    if tool_name == "edit_file":
        if physical_path is None:
            return ApprovalPreview(
                title=f"Update {display_path}",
                details=[f"File: {path_str}", "Action: Replace text"],
                error="Unable to resolve file path.",
            )
        before = _safe_read(physical_path)
        if before is None:
            return ApprovalPreview(
                title=f"Update {display_path}",
                details=[f"File: {path_str}", "Action: Replace text"],
                error="Unable to read current file contents.",
            )
        old_string = str(args.get("old_string", ""))
        new_string = str(args.get("new_string", ""))
        replace_all = bool(args.get("replace_all"))
        from deepagents.backends.utils import perform_string_replacement

        replacement = perform_string_replacement(
            before, old_string, new_string, replace_all
        )
        if isinstance(replacement, str):
            return ApprovalPreview(
                title=f"Update {display_path}",
                details=[f"File: {path_str}", "Action: Replace text"],
                error=replacement,
            )
        after, occurrences = replacement
        diff = compute_unified_diff(before, after, display_path, max_lines=None)
        additions = 0
        deletions = 0
        if diff:
            additions = sum(
                1
                for line in diff.splitlines()
                if line.startswith("+") and not line.startswith("+++")
            )
            deletions = sum(
                1
                for line in diff.splitlines()
                if line.startswith("-") and not line.startswith("---")
            )
        action = "all occurrences" if replace_all else "single occurrence"
        details = [
            f"File: {path_str}",
            f"Action: Replace text ({action})",
            f"Occurrences matched: {occurrences}",
            f"Lines changed: +{additions} / -{deletions}",
        ]
        return ApprovalPreview(
            title=f"Update {display_path}",
            details=details,
            diff=diff,
            diff_title=f"Diff {display_path}",
        )

    return None


class FileOpTracker:
    """Collect file operation metrics during a CLI interaction."""

    def __init__(
        self, *, assistant_id: str | None, backend: BackendProtocol | None = None
    ) -> None:
        """Initialize the tracker."""
        self.assistant_id = assistant_id
        self.backend = backend
        self.active: dict[str | None, FileOperationRecord] = {}
        self.completed: list[FileOperationRecord] = []

    def start_operation(
        self, tool_name: str, args: dict[str, Any], tool_call_id: str | None
    ) -> None:
        """Begin tracking a file operation.

        Creates a record for the operation and, for write/edit operations,
        captures the file's content before modification.
        """
        if tool_name not in {"read_file", "write_file", "edit_file"}:
            return
        path_str = str(args.get("file_path") or args.get("path") or "")
        display_path = format_display_path(path_str)
        record = FileOperationRecord(
            tool_name=tool_name,
            display_path=display_path,
            physical_path=resolve_physical_path(path_str, self.assistant_id),
            tool_call_id=tool_call_id,
            args=args,
        )
        if tool_name in {"write_file", "edit_file"}:
            if self.backend and path_str:
                try:
                    responses = self.backend.download_files([path_str])
                    if (
                        responses
                        and responses[0].content is not None
                        and responses[0].error is None
                    ):
                        record.before_content = responses[0].content.decode("utf-8")
                    else:
                        record.before_content = ""
                except (OSError, UnicodeDecodeError, AttributeError) as e:
                    logger.debug(
                        "Failed to read before_content for %s: %s", path_str, e
                    )
                    record.before_content = ""
            elif record.physical_path:
                record.before_content = _safe_read(record.physical_path) or ""
        self.active[tool_call_id] = record

    def complete_with_message(self, tool_message: Any) -> FileOperationRecord | None:  # noqa: ANN401  # Tool message type is dynamic
        """Complete a file operation with the tool message result.

        Returns:
            The completed FileOperationRecord, or None if no matching operation.
        """
        tool_call_id = getattr(tool_message, "tool_call_id", None)
        record = self.active.get(tool_call_id)
        if record is None:
            return None

        content = tool_message.content
        if isinstance(content, list):
            # Some tool messages may return list segments; join them for analysis.
            joined = []
            for item in content:
                if isinstance(item, str):
                    joined.append(item)
                else:
                    joined.append(str(item))
            content_text = "\n".join(joined)
        else:
            content_text = str(content) if content is not None else ""

        if getattr(
            tool_message, "status", "success"
        ) != "success" or content_text.lower().startswith("error"):
            record.status = "error"
            record.error = content_text
            self._finalize(record)
            return record

        record.status = "success"

        if record.tool_name == "read_file":
            record.read_output = content_text
            lines = _count_lines(content_text)
            record.metrics.lines_read = lines
            offset = record.args.get("offset")
            limit = record.args.get("limit")
            if isinstance(offset, int):
                if offset > lines:
                    offset = 0
                record.metrics.start_line = offset + 1
                if lines:
                    record.metrics.end_line = offset + lines
            elif lines:
                record.metrics.start_line = 1
                record.metrics.end_line = lines
            if isinstance(limit, int) and lines > limit:
                record.metrics.end_line = (record.metrics.start_line or 1) + limit - 1
        else:
            # For write/edit operations, read back from backend (or local filesystem)
            self._populate_after_content(record)
            if record.after_content is None:
                record.status = "error"
                record.error = "Could not read updated file content."
                self._finalize(record)
                return record
            record.metrics.lines_written = _count_lines(record.after_content)
            before_lines = _count_lines(record.before_content or "")
            diff = compute_unified_diff(
                record.before_content or "",
                record.after_content,
                record.display_path,
                max_lines=100,
            )
            record.diff = diff
            if diff:
                additions = sum(
                    1
                    for line in diff.splitlines()
                    if line.startswith("+") and not line.startswith("+++")
                )
                deletions = sum(
                    1
                    for line in diff.splitlines()
                    if line.startswith("-") and not line.startswith("---")
                )
                record.metrics.lines_added = additions
                record.metrics.lines_removed = deletions
            elif record.tool_name == "write_file" and not (record.before_content or ""):
                record.metrics.lines_added = record.metrics.lines_written
            record.metrics.bytes_written = len(record.after_content.encode("utf-8"))
            if (
                record.diff is None
                and (record.before_content or "") != record.after_content
            ):
                record.diff = compute_unified_diff(
                    record.before_content or "",
                    record.after_content,
                    record.display_path,
                    max_lines=100,
                )
            if record.diff is None and before_lines != record.metrics.lines_written:
                record.metrics.lines_added = max(
                    record.metrics.lines_written - before_lines, 0
                )

        self._finalize(record)
        return record

    def mark_hitl_approved(self, tool_name: str, args: dict[str, Any]) -> None:
        """Mark operations matching tool_name and file_path as HIL-approved."""
        file_path = args.get("file_path") or args.get("path")
        if not file_path:
            return

        # Mark all active records that match
        for record in self.active.values():
            if record.tool_name == tool_name:
                record_path = record.args.get("file_path") or record.args.get("path")
                if record_path == file_path:
                    record.hitl_approved = True

    def _populate_after_content(self, record: FileOperationRecord) -> None:
        # Use backend if available (works for any BackendProtocol implementation)
        if self.backend:
            try:
                file_path = record.args.get("file_path") or record.args.get("path")
                if file_path:
                    responses = self.backend.download_files([file_path])
                    if (
                        responses
                        and responses[0].content is not None
                        and responses[0].error is None
                    ):
                        record.after_content = responses[0].content.decode("utf-8")
                    else:
                        record.after_content = None
                else:
                    record.after_content = None
            except (OSError, UnicodeDecodeError, AttributeError) as e:
                logger.debug(
                    "Failed to read after_content for %s: %s",
                    record.args.get("file_path") or record.args.get("path"),
                    e,
                )
                record.after_content = None
        else:
            # Fallback: direct filesystem read when no backend provided
            if record.physical_path is None:
                record.after_content = None
                return
            record.after_content = _safe_read(record.physical_path)

    def _finalize(self, record: FileOperationRecord) -> None:
        self.completed.append(record)
        self.active.pop(record.tool_call_id, None)



"""Agent management and creation for the CLI."""

from __future__ import annotations

import logging
import os
import re
import shutil
import tempfile
import tomllib
from pathlib import Path
from typing import TYPE_CHECKING, Any

from deepagents import create_deep_agent
from deepagents.backends import CompositeBackend, LocalShellBackend
from deepagents.backends.filesystem import FilesystemBackend
from deepagents.middleware import MemoryMiddleware, SkillsMiddleware

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from deepagents.backends.sandbox import SandboxBackendProtocol
    from deepagents.middleware.async_subagents import AsyncSubAgent
    from deepagents.middleware.subagents import CompiledSubAgent, SubAgent
    from langchain.agents.middleware import InterruptOnConfig
    from langchain.agents.middleware.types import AgentState
    from langchain.messages import ToolCall
    from langchain.tools import BaseTool
    from langchain_core.language_models import BaseChatModel
    from langgraph.checkpoint.base import BaseCheckpointSaver
    from langgraph.pregel import Pregel
    from langgraph.runtime import Runtime

    from deepagents_cli.mcp_tools import MCPServerInfo
    from deepagents_cli.output import OutputFormat

from deepagents_cli.config import (
    COLORS,
    config,
    console,
    get_default_coding_instructions,
    get_glyphs,
    settings,
)
from deepagents_cli.configurable_model import ConfigurableModelMiddleware
from deepagents_cli.integrations.sandbox_factory import get_default_working_dir
from deepagents_cli.local_context import LocalContextMiddleware, _ExecutableBackend
from deepagents_cli.project_utils import ProjectContext, get_server_project_context
from deepagents_cli.subagents import list_subagents
from deepagents_cli.unicode_security import (
    check_url_safety,
    detect_dangerous_unicode,
    format_warning_detail,
    render_with_unicode_markers,
    strip_dangerous_unicode,
    summarize_issues,
)

logger = logging.getLogger(__name__)

DEFAULT_AGENT_NAME = "agent"
"""The default agent name used when no `-a` flag is provided."""

REQUIRE_COMPACT_TOOL_APPROVAL: bool = True
"""When `True`, `compact_conversation` requires HITL approval like other gated tools."""


def load_async_subagents(config_path: Path | None = None) -> list[AsyncSubAgent]:
    """Load async subagent definitions from `config.toml`.

    Reads the `[async_subagents]` section where each sub-table defines a remote
    LangGraph deployment:

    ```toml
    [async_subagents.researcher]
    description = "Research agent"
    url = "https://my-deployment.langsmith.dev"
    graph_id = "agent"
    ```

    Args:
        config_path: Path to config file.

            Defaults to `~/.deepagents/config.toml`.

    Returns:
        List of `AsyncSubAgent` specs (empty if section is absent or invalid).
    """
    if config_path is None:
        config_path = Path.home() / ".deepagents" / "config.toml"

    if not config_path.exists():
        return []

    try:
        with config_path.open("rb") as f:
            data = tomllib.load(f)
    except (tomllib.TOMLDecodeError, PermissionError, OSError) as e:
        logger.warning("Could not read async subagents from %s: %s", config_path, e)
        console.print(
            f"[bold yellow]Warning:[/bold yellow] Could not read async subagents "
            f"from {config_path}: {e}",
        )
        return []

    section = data.get("async_subagents")
    if not isinstance(section, dict):
        return []

    required = {"description", "graph_id"}
    agents: list[AsyncSubAgent] = []
    for name, spec in section.items():
        if not isinstance(spec, dict):
            logger.warning("Skipping async subagent '%s': expected a table", name)
            continue
        missing = required - spec.keys()
        if missing:
            logger.warning(
                "Skipping async subagent '%s': missing fields %s", name, missing
            )
            continue
        agent: AsyncSubAgent = {
            "name": name,
            "description": spec["description"],
            "graph_id": spec["graph_id"],
        }
        if "url" in spec and isinstance(spec["url"], str):
            agent["url"] = spec["url"]
        if "headers" in spec and isinstance(spec["headers"], dict):
            agent["headers"] = spec["headers"]
        agents.append(agent)

    return agents


def list_agents(*, output_format: OutputFormat = "text") -> None:
    """List all available agents.

    Args:
        output_format: Output format — `'text'` (Rich) or `'json'`.
    """
    agents_dir = settings.user_deepagents_dir

    if not agents_dir.exists() or not any(agents_dir.iterdir()):
        if output_format == "json":
            from deepagents_cli.output import write_json

            write_json("list", [])
            return
        console.print("[yellow]No agents found.[/yellow]")
        console.print(
            "[dim]Agents will be created in ~/.deepagents/ "
            "when you first use them.[/dim]",
            style=COLORS["dim"],
        )
        return

    if output_format == "json":
        from deepagents_cli.output import write_json

        agents = []
        for agent_path in sorted(agents_dir.iterdir()):
            if agent_path.is_dir():
                agent_name = agent_path.name
                agents.append(
                    {
                        "name": agent_name,
                        "path": str(agent_path),
                        "has_agents_md": (agent_path / "AGENTS.md").exists(),
                        "is_default": agent_name == DEFAULT_AGENT_NAME,
                    }
                )
        write_json("list", agents)
        return

    from rich.markup import escape as escape_markup

    console.print("\n[bold]Available Agents:[/bold]\n", style=COLORS["primary"])

    for agent_path in sorted(agents_dir.iterdir()):
        if agent_path.is_dir():
            agent_name = escape_markup(agent_path.name)
            agent_md = agent_path / "AGENTS.md"
            is_default = agent_path.name == DEFAULT_AGENT_NAME
            default_label = " [dim](default)[/dim]" if is_default else ""

            bullet = get_glyphs().bullet
            if agent_md.exists():
                console.print(
                    f"  {bullet} [bold]{agent_name}[/bold]{default_label}",
                    style=COLORS["primary"],
                )
                console.print(
                    f"    {escape_markup(str(agent_path))}",
                    style=COLORS["dim"],
                )
            else:
                console.print(
                    f"  {bullet} [bold]{agent_name}[/bold]{default_label}"
                    " [dim](incomplete)[/dim]",
                    style=COLORS["tool"],
                )
                console.print(
                    f"    {escape_markup(str(agent_path))}",
                    style=COLORS["dim"],
                )

    console.print()


def reset_agent(
    agent_name: str,
    source_agent: str | None = None,
    *,
    output_format: OutputFormat = "text",
) -> None:
    """Reset an agent to default or copy from another agent.

    Args:
        agent_name: Name of the agent to reset.
        source_agent: Copy AGENTS.md from this agent instead of default.
        output_format: Output format — `'text'` (Rich) or `'json'`.
    """
    agents_dir = settings.user_deepagents_dir
    agent_dir = agents_dir / agent_name

    if source_agent:
        source_dir = agents_dir / source_agent
        source_md = source_dir / "AGENTS.md"

        if not source_md.exists():
            console.print(
                f"[bold red]Error:[/bold red] Source agent '{source_agent}' not found "
                "or has no AGENTS.md"
            )
            return

        source_content = source_md.read_text()
        action_desc = f"contents of agent '{source_agent}'"
    else:
        source_content = get_default_coding_instructions()
        action_desc = "default"

    if agent_dir.exists():
        shutil.rmtree(agent_dir)
        if output_format != "json":
            console.print(
                f"Removed existing agent directory: {agent_dir}", style=COLORS["tool"]
            )

    agent_dir.mkdir(parents=True, exist_ok=True)
    agent_md = agent_dir / "AGENTS.md"
    agent_md.write_text(source_content)

    if output_format == "json":
        from deepagents_cli.output import write_json

        write_json(
            "reset",
            {
                "agent": agent_name,
                "reset_to": source_agent or "default",
                "path": str(agent_dir),
            },
        )
        return

    console.print(
        f"{get_glyphs().checkmark} Agent '{agent_name}' reset to {action_desc}",
        style=COLORS["primary"],
    )
    console.print(f"Location: {agent_dir}\n", style=COLORS["dim"])


MODEL_IDENTITY_RE = re.compile(r"### Model Identity\n\n.*?(?=###|\Z)", re.DOTALL)
"""Matches the `### Model Identity` section in the system prompt, up to the
next heading or end of string."""


def build_model_identity_section(
    name: str | None,
    provider: str | None = None,
    context_limit: int | None = None,
) -> str:
    """Build the `### Model Identity` section for the system prompt.

    Args:
        name: Model identifier (e.g. `claude-opus-4-6`).
        provider: Provider identifier (e.g. `anthropic`).
        context_limit: Max input tokens from the model profile.

    Returns:
        The section text including the heading and trailing newline,
        or an empty string if `name` is falsy.
    """
    if not name:
        return ""
    section = f"### Model Identity\n\nYou are running as model `{name}`"
    if provider:
        section += f" (provider: {provider})"
    section += ".\n"
    if context_limit:
        section += f"Your context window is {context_limit:,} tokens.\n"
    section += "\n"
    return section


def get_system_prompt(
    assistant_id: str,
    sandbox_type: str | None = None,
    *,
    interactive: bool = True,
    cwd: str | Path | None = None,
) -> str:
    """Get the base system prompt for the agent.

    Loads the base system prompt template from `system_prompt.md` and
    interpolates dynamic sections (model identity, working directory,
    skills path, execution mode).

    Args:
        assistant_id: The agent identifier for path references
        sandbox_type: Type of sandbox provider
            (`'daytona'`, `'langsmith'`, `'modal'`, `'runloop'`).

            If `None`, agent is operating in local mode.
        interactive: When `False`, the prompt is tailored for headless
            non-interactive execution (no human in the loop).
        cwd: Override the working directory shown in the prompt.

    Returns:
        The system prompt string

    Example:
        ```txt
        You are running as model {MODEL} (provider: {PROVIDER}).

        Your context window is {CONTEXT_WINDOW} tokens.

        ... {CONDITIONAL SECTIONS} ...
        ```
    """
    template = (Path(__file__).parent / "system_prompt.md").read_text()

    skills_path = f"~/.deepagents/{assistant_id}/skills"

    if interactive:
        mode_description = "an interactive CLI on the user's computer"
        interactive_preamble = (
            "The user sends you messages and you respond with text and tool "
            "calls. Your tools run on the user's machine. The user can see "
            "your responses and tool outputs in real time, so keep them "
            "informed — but don't over-explain."
        )
        ambiguity_guidance = (
            "- If the request is ambiguous, ask questions before acting.\n"
            "- If asked how to approach something, explain first, then act."
        )
    else:
        mode_description = (
            "non-interactive (headless) mode — there is no human operator "
            "monitoring your output in real time"
        )
        interactive_preamble = (
            "You received a single task and must complete it fully and "
            "autonomously. There is no human available to answer follow-up "
            "questions, so do NOT ask for clarification — make reasonable "
            "assumptions and proceed."
        )
        ambiguity_guidance = (
            "- Do NOT ask clarifying questions — there is no human to answer "
            "them. Make reasonable assumptions and proceed.\n"
            "- If you encounter ambiguity, choose the most reasonable "
            "interpretation and note your assumption briefly.\n"
            "- Always use non-interactive command variants — no human is "
            "available to respond to prompts. Examples: `npm init -y` not "
            "`npm init`, `apt-get install -y` not `apt-get install`, "
            "`yes |` or `--no-input`/`--non-interactive` flags where "
            "available. Never run commands that block waiting for stdin."
        )

    model_identity_section = build_model_identity_section(
        settings.model_name,
        provider=settings.model_provider,
        context_limit=settings.model_context_limit,
    )

    # Build working directory section (local vs sandbox)
    if sandbox_type:
        working_dir = get_default_working_dir(sandbox_type)
        working_dir_section = (
            f"### Current Working Directory\n\n"
            f"You are operating in a **remote Linux sandbox** at `{working_dir}`.\n\n"
            f"All code execution and file operations happen in this sandbox "
            f"environment.\n\n"
            f"**Important:**\n"
            f"- The CLI is running locally on the user's machine, but you execute "
            f"code remotely\n"
            f"- Use `{working_dir}` as your working directory for all operations\n\n"
        )
    else:
        if cwd is not None:
            resolved_cwd = Path(cwd)
        else:
            try:
                resolved_cwd = Path.cwd()
            except OSError:
                logger.warning(
                    "Could not determine working directory for system prompt",
                    exc_info=True,
                )
                resolved_cwd = Path()
        cwd = resolved_cwd
        working_dir_section = (
            f"### Current Working Directory\n\n"
            f"The filesystem backend is currently operating in: `{cwd}`\n\n"
            f"### File System and Paths\n\n"
            f"**IMPORTANT - Path Handling:**\n"
            f"- All file paths must be absolute paths (e.g., `{cwd}/file.txt`)\n"
            f"- Use the working directory to construct absolute paths\n"
            f"- Example: To create a file in your working directory, "
            f"use `{cwd}/research_project/file.md`\n"
            f"- Never use relative paths - always construct full absolute paths\n\n"
        )

    result = (
        template.replace("{mode_description}", mode_description)
        .replace("{interactive_preamble}", interactive_preamble)
        .replace("{ambiguity_guidance}", ambiguity_guidance)
        .replace("{model_identity_section}", model_identity_section)
        .replace("{working_dir_section}", working_dir_section)
        .replace("{skills_path}", skills_path)
    )

    # CXCoast org customization: inject org name, custom prompt, and instructions
    import json as _json
    _cxcoast_config_path = Path.home() / ".cxcoast" / "config.json"
    if _cxcoast_config_path.exists():
        try:
            _cx_config = _json.loads(_cxcoast_config_path.read_text())
            _org_name = _cx_config.get("orgName", "CXCoast")
            _custom_prompt = _cx_config.get("customSystemPrompt")
            _org_instructions = _cx_config.get("orgInstructions")
            _default_language = _cx_config.get("defaultLanguage")

            # CXCoast default system prompt (same as agent-service prompts.py)
            _CXCOAST_DEFAULT_PROMPT = (
                "You are {orgName}'s Knowledge Assistant, powered by the CXCoast Knowledge Hub.\n"
                "Do NOT identify as \"Deep Agent\" or reference the DeepAgents framework.\n\n"
                "## Your Tools\n\n"
                "- **search_knowledge**: Search the knowledge base (hybrid search — semantic + keyword)\n"
                "- **get_article**: Retrieve a specific article with audience-appropriate content\n"
                "- **list_categories**: Browse the knowledge taxonomy\n"
                "- **query_database**: Get statistics (article count, gaps, usage, freshness)\n\n"
                "## How to Work\n\n"
                "- ALWAYS search before answering knowledge questions — never guess or hallucinate\n"
                "- Cite your sources: include article titles and IDs in your answers\n"
                "- For complex questions, break them into sub-tasks using the planning system\n"
                "- If the knowledge base doesn't have the answer, say so clearly and suggest what content should be created\n"
                "- When showing search results, summarize — don't dump raw JSON\n"
            )

            # Apply custom prompt if set, otherwise use CXCoast default
            _template = _custom_prompt if _custom_prompt else _CXCOAST_DEFAULT_PROMPT
            # Inject orgName
            _resolved = _template.replace("{orgName}", _org_name)
            # Prepend resolved CXCoast prompt to the SDK prompt (SDK handles tools, file ops, etc.)
            result = _resolved + "\n\n" + result

            # Append org instructions (like CLAUDE.md)
            if _org_instructions:
                result += "\n\n--- Organization Instructions ---\n" + _org_instructions

            # Append language preference
            if _default_language:
                result += f"\n\nPreferred response language: {_default_language}"
        except Exception:
            pass

    # Detect unreplaced placeholders (defense-in-depth for template typos)
    unreplaced = re.findall(r"\{[a-z_]+\}", result)
    if unreplaced:
        logger.warning("System prompt contains unreplaced placeholders: %s", unreplaced)

    return result


def _format_write_file_description(
    tool_call: ToolCall, _state: AgentState[Any], _runtime: Runtime[Any]
) -> str:
    """Format write_file tool call for approval prompt.

    Returns:
        Formatted description string for the write_file tool call.
    """
    args = tool_call["args"]
    file_path = args.get("file_path", "unknown")
    content = args.get("content", "")

    action = "Overwrite" if Path(file_path).exists() else "Create"
    line_count = len(content.splitlines())

    return f"File: {file_path}\nAction: {action} file\nLines: {line_count}"


def _format_edit_file_description(
    tool_call: ToolCall, _state: AgentState[Any], _runtime: Runtime[Any]
) -> str:
    """Format edit_file tool call for approval prompt.

    Returns:
        Formatted description string for the edit_file tool call.
    """
    args = tool_call["args"]
    file_path = args.get("file_path", "unknown")
    replace_all = bool(args.get("replace_all", False))

    scope = "all occurrences" if replace_all else "single occurrence"
    return f"File: {file_path}\nAction: Replace text ({scope})"


def _format_web_search_description(
    tool_call: ToolCall, _state: AgentState[Any], _runtime: Runtime[Any]
) -> str:
    """Format web_search tool call for approval prompt.

    Returns:
        Formatted description string for the web_search tool call.
    """
    args = tool_call["args"]
    query = args.get("query", "unknown")
    max_results = args.get("max_results", 5)

    return (
        f"Query: {query}\nMax results: {max_results}\n\n"
        f"{get_glyphs().warning}  This will use Tavily API credits"
    )


def _format_fetch_url_description(
    tool_call: ToolCall, _state: AgentState[Any], _runtime: Runtime[Any]
) -> str:
    """Format fetch_url tool call for approval prompt.

    Returns:
        Formatted description string for the fetch_url tool call.
    """
    args = tool_call["args"]
    url = str(args.get("url", "unknown"))
    display_url = strip_dangerous_unicode(url)
    timeout = args.get("timeout", 30)
    safety = check_url_safety(url)

    warning_lines: list[str] = []
    if not safety.safe:
        detail = format_warning_detail(safety.warnings)
        warning_lines.append(f"{get_glyphs().warning}  URL warning: {detail}")
    if safety.decoded_domain:
        warning_lines.append(
            f"{get_glyphs().warning}  Decoded domain: {safety.decoded_domain}"
        )

    warning_block = "\n".join(warning_lines)
    if warning_block:
        warning_block = f"\n{warning_block}"

    return (
        f"URL: {display_url}\nTimeout: {timeout}s\n\n"
        f"{get_glyphs().warning}  Will fetch and convert web content to markdown"
        f"{warning_block}"
    )


def _format_task_description(
    tool_call: ToolCall, _state: AgentState[Any], _runtime: Runtime[Any]
) -> str:
    """Format task (subagent) tool call for approval prompt.

    The task tool signature is: task(description: str, subagent_type: str)
    The description contains all instructions that will be sent to the subagent.

    Returns:
        Formatted description string for the task tool call.
    """
    args = tool_call["args"]
    description = args.get("description", "unknown")
    subagent_type = args.get("subagent_type", "unknown")

    # Truncate description if too long for display
    description_preview = description
    if len(description) > 500:  # noqa: PLR2004  # Subagent description length threshold
        description_preview = description[:500] + "..."

    glyphs = get_glyphs()
    separator = glyphs.box_horizontal * 40
    warning_msg = "Subagent will have access to file operations and shell commands"
    return (
        f"Subagent Type: {subagent_type}\n\n"
        f"Task Instructions:\n"
        f"{separator}\n"
        f"{description_preview}\n"
        f"{separator}\n\n"
        f"{glyphs.warning}  {warning_msg}"
    )


def _format_execute_description(
    tool_call: ToolCall, _state: AgentState[Any], _runtime: Runtime[Any]
) -> str:
    """Format execute tool call for approval prompt.

    Returns:
        Formatted description string for the execute tool call.
    """
    args = tool_call["args"]
    command_raw = str(args.get("command", "N/A"))
    command = strip_dangerous_unicode(command_raw)
    project_context = get_server_project_context()
    effective_cwd = (
        str(project_context.user_cwd)
        if project_context is not None
        else str(Path.cwd())
    )
    lines = [f"Execute Command: {command}", f"Working Directory: {effective_cwd}"]

    issues = detect_dangerous_unicode(command_raw)
    if issues:
        summary = summarize_issues(issues)
        lines.append(f"{get_glyphs().warning}  Hidden Unicode detected: {summary}")
        raw_marked = render_with_unicode_markers(command_raw)
        if len(raw_marked) > 220:  # noqa: PLR2004  # UI display truncation threshold
            raw_marked = raw_marked[:220] + "..."
        lines.append(f"Raw: {raw_marked}")

    return "\n".join(lines)


def _add_interrupt_on() -> dict[str, InterruptOnConfig]:
    """Configure human-in-the-loop interrupt settings for all gated tools.

    Every tool that can have side effects or access external resources
    (shell execution, file writes/edits, web search, URL fetch, task
    delegation) is gated behind an approval prompt unless auto-approve
    is enabled.

    Returns:
        Dictionary mapping tool names to their interrupt configuration.
    """
    execute_interrupt_config: InterruptOnConfig = {
        "allowed_decisions": ["approve", "reject"],
        "description": _format_execute_description,  # type: ignore[typeddict-item]  # Callable description narrower than TypedDict expects
    }

    write_file_interrupt_config: InterruptOnConfig = {
        "allowed_decisions": ["approve", "reject"],
        "description": _format_write_file_description,  # type: ignore[typeddict-item]  # Callable description narrower than TypedDict expects
    }

    edit_file_interrupt_config: InterruptOnConfig = {
        "allowed_decisions": ["approve", "reject"],
        "description": _format_edit_file_description,  # type: ignore[typeddict-item]  # Callable description narrower than TypedDict expects
    }

    web_search_interrupt_config: InterruptOnConfig = {
        "allowed_decisions": ["approve", "reject"],
        "description": _format_web_search_description,  # type: ignore[typeddict-item]  # Callable description narrower than TypedDict expects
    }

    fetch_url_interrupt_config: InterruptOnConfig = {
        "allowed_decisions": ["approve", "reject"],
        "description": _format_fetch_url_description,  # type: ignore[typeddict-item]  # Callable description narrower than TypedDict expects
    }

    task_interrupt_config: InterruptOnConfig = {
        "allowed_decisions": ["approve", "reject"],
        "description": _format_task_description,  # type: ignore[typeddict-item]  # Callable description narrower than TypedDict expects
    }

    async_subagent_interrupt_config: InterruptOnConfig = {
        "allowed_decisions": ["approve", "reject"],
        "description": "Launch, update, or cancel a remote async subagent.",
    }

    interrupt_map: dict[str, InterruptOnConfig] = {
        "execute": execute_interrupt_config,
        "write_file": write_file_interrupt_config,
        "edit_file": edit_file_interrupt_config,
        "web_search": web_search_interrupt_config,
        "fetch_url": fetch_url_interrupt_config,
        "task": task_interrupt_config,
        "launch_async_subagent": async_subagent_interrupt_config,
        "update_async_subagent": async_subagent_interrupt_config,
        "cancel_async_subagent": async_subagent_interrupt_config,
    }

    if REQUIRE_COMPACT_TOOL_APPROVAL:
        interrupt_map["compact_conversation"] = {
            "allowed_decisions": ["approve", "reject"],
            "description": (
                "Offloads older messages to backend storage and "
                "replaces them with a summary, freeing context "
                "window space. Recent messages are kept as-is. "
                "Full history remains available for retrieval."
            ),
        }

    return interrupt_map


def create_cli_agent(
    model: str | BaseChatModel,
    assistant_id: str,
    *,
    tools: Sequence[BaseTool | Callable | dict[str, Any]] | None = None,
    sandbox: SandboxBackendProtocol | None = None,
    sandbox_type: str | None = None,
    system_prompt: str | None = None,
    interactive: bool = True,
    auto_approve: bool = False,
    enable_ask_user: bool = True,
    enable_memory: bool = True,
    enable_skills: bool = True,
    enable_shell: bool = True,
    checkpointer: BaseCheckpointSaver | None = None,
    mcp_server_info: list[MCPServerInfo] | None = None,
    cwd: str | Path | None = None,
    project_context: ProjectContext | None = None,
    async_subagents: list[AsyncSubAgent] | None = None,
) -> tuple[Pregel, CompositeBackend]:
    """Create a CLI-configured agent with flexible options.

    This is the main entry point for creating a deepagents CLI agent, usable
    both internally and from external code (e.g., benchmarking frameworks).

    Args:
        model: LLM model to use (e.g., `'anthropic:claude-sonnet-4-6'`)
        assistant_id: Agent identifier for memory/state storage
        tools: Additional tools to provide to agent
        sandbox: Optional sandbox backend for remote execution
            (e.g., `ModalSandbox`).

            If `None`, uses local filesystem + shell.
        sandbox_type: Type of sandbox provider
            (`'daytona'`, `'langsmith'`, `'modal'`, `'runloop'`).
            Used for system prompt generation.
        system_prompt: Override the default system prompt.

            If `None`, generates one based on `sandbox_type`, `assistant_id`,
            and `interactive`.
        interactive: When `False`, the auto-generated system prompt is
            tailored for headless non-interactive execution. Ignored when
            `system_prompt` is provided explicitly.
        auto_approve: If `True`, no tools trigger human-in-the-loop
            interrupts — all calls (shell execution, file writes/edits,
            web search, URL fetch) run automatically.

            If `False`, tools pause for user confirmation via the approval menu.
            See `_add_interrupt_on` for the full list of gated tools.
        enable_ask_user: Enable `AskUserMiddleware` so the agent can ask
            clarifying questions.

            Disabled in non-interactive mode.
        enable_memory: Enable `MemoryMiddleware` for persistent memory
        enable_skills: Enable `SkillsMiddleware` for custom agent skills
        enable_shell: Enable shell execution via `LocalShellBackend`
            (only in local mode). When enabled, the `execute` tool is available.
        checkpointer: Optional checkpointer for session persistence.
            When `None`, the graph is compiled without a checkpointer.
        mcp_server_info: MCP server metadata to surface in the system prompt.
        cwd: Override the working directory for the agent's filesystem backend
            and system prompt.
        project_context: Explicit project path context for project-sensitive
            behavior such as project `AGENTS.md` files, skills, subagents, and
            MCP trust.
        async_subagents: Remote LangGraph deployments to expose as async subagent tools.

            Loaded from `[async_subagents]` in `config.toml` or passed directly.

    Returns:
        2-tuple of `(agent_graph, backend)`

            - `agent_graph`: Configured LangGraph Pregel instance ready
                for execution
            - `composite_backend`: `CompositeBackend` for file operations
    """
    tools = tools or []
    effective_cwd = (
        Path(cwd)
        if cwd is not None
        else (project_context.user_cwd if project_context is not None else None)
    )

    # Setup agent directory for persistent memory (if enabled)
    if enable_memory or enable_skills:
        agent_dir = settings.ensure_agent_dir(assistant_id)
        agent_md = agent_dir / "AGENTS.md"
        if not agent_md.exists():
            # Create empty file for user customizations
            # Base instructions are loaded fresh from get_system_prompt()
            agent_md.touch()

    # Skills directories (if enabled)
    skills_dir = None
    user_agent_skills_dir = None
    project_skills_dir = None
    project_agent_skills_dir = None
    if enable_skills:
        skills_dir = settings.ensure_user_skills_dir(assistant_id)
        user_agent_skills_dir = settings.get_user_agent_skills_dir()
        project_skills_dir = (
            project_context.project_skills_dir()
            if project_context is not None
            else settings.get_project_skills_dir()
        )
        project_agent_skills_dir = (
            project_context.project_agent_skills_dir()
            if project_context is not None
            else settings.get_project_agent_skills_dir()
        )

    # Load custom subagents from filesystem
    custom_subagents: list[SubAgent | CompiledSubAgent] = []
    user_agents_dir = settings.get_user_agents_dir(assistant_id)
    project_agents_dir = (
        project_context.project_agents_dir()
        if project_context is not None
        else settings.get_project_agents_dir()
    )

    for subagent_meta in list_subagents(
        user_agents_dir=user_agents_dir,
        project_agents_dir=project_agents_dir,
    ):
        subagent: SubAgent = {
            "name": subagent_meta["name"],
            "description": subagent_meta["description"],
            "system_prompt": subagent_meta["system_prompt"],
        }
        if subagent_meta["model"]:
            subagent["model"] = subagent_meta["model"]
        custom_subagents.append(subagent)

    # Build middleware stack based on enabled features
    agent_middleware = []
    agent_middleware.append(ConfigurableModelMiddleware())

    # Add ask_user middleware (must be early so its tool is available)
    if enable_ask_user:
        from deepagents_cli.ask_user import AskUserMiddleware

        agent_middleware.append(AskUserMiddleware())

    # Add memory middleware
    if enable_memory:
        memory_sources = [str(settings.get_user_agent_md_path(assistant_id))]
        project_agent_md_paths = (
            project_context.project_agent_md_paths()
            if project_context is not None
            else settings.get_project_agent_md_path()
        )
        memory_sources.extend(str(p) for p in project_agent_md_paths)

        agent_middleware.append(
            MemoryMiddleware(
                backend=FilesystemBackend(),
                sources=memory_sources,
            )
        )

    # Add skills middleware
    if enable_skills:
        _skills_backend = FilesystemBackend()
        _skills_sources = [str(settings.get_built_in_skills_dir())]
        _skills_sources.extend([str(skills_dir), str(user_agent_skills_dir)])
        if project_skills_dir:
            _skills_sources.append(str(project_skills_dir))
        if project_agent_skills_dir:
            _skills_sources.append(str(project_agent_skills_dir))

        # Add CXCoast platform skills (downloaded to ~/.cxcoast/skills/)
        _cxcoast_skills_dir = Path.home() / ".cxcoast" / "skills"
        if _cxcoast_skills_dir.exists() and any(
            p.is_dir() for p in _cxcoast_skills_dir.iterdir() if p.name != ".version"
        ):
            _skills_sources.append(str(_cxcoast_skills_dir))

        agent_middleware.append(
            SkillsMiddleware(
                backend=_skills_backend,
                sources=_skills_sources,
            )
        )

    # CONDITIONAL SETUP: Local vs Remote Sandbox
    if sandbox is None:
        # ========== LOCAL MODE ==========
        root_dir = effective_cwd if effective_cwd is not None else Path.cwd()
        if enable_shell:
            # Create environment for shell commands
            # Restore user's original LANGSMITH_PROJECT so their code traces separately
            shell_env = os.environ.copy()
            if settings.user_langchain_project:
                shell_env["LANGSMITH_PROJECT"] = settings.user_langchain_project

            # Use LocalShellBackend for filesystem + shell execution.
            # The SDK's FilesystemMiddleware exposes per-command timeout
            # on the execute tool natively.
            backend = LocalShellBackend(
                root_dir=root_dir,
                inherit_env=True,
                env=shell_env,
            )
        else:
            # No shell access - use plain FilesystemBackend
            backend = FilesystemBackend(root_dir=root_dir)
    else:
        # ========== REMOTE SANDBOX MODE ==========
        backend = sandbox  # Remote sandbox (ModalSandbox, etc.)
        # Note: Shell middleware not used in sandbox mode
        # File operations and execute tool are provided by the sandbox backend

    # Local context middleware (git info, directory tree, etc.)
    # Uses backend.execute() so it works in both local shell and remote sandbox modes.
    # Only enabled when the backend supports shell execution.
    if isinstance(backend, _ExecutableBackend):
        agent_middleware.append(
            LocalContextMiddleware(backend=backend, mcp_server_info=mcp_server_info)
        )

    # Get or use custom system prompt
    if system_prompt is None:
        system_prompt = get_system_prompt(
            assistant_id=assistant_id,
            sandbox_type=sandbox_type,
            interactive=interactive,
            cwd=effective_cwd,
        )

    # Configure interrupt_on based on auto_approve setting
    interrupt_on: dict[str, bool | InterruptOnConfig] | None = None
    if auto_approve:  # noqa: SIM108  # if-else more readable for interrupt_on config
        # No interrupts - all tools run automatically
        interrupt_on = {}
    else:
        # Full HITL for destructive operations
        interrupt_on = _add_interrupt_on()  # type: ignore[assignment]  # InterruptOnConfig is compatible at runtime

    # Set up composite backend with routing
    # For local FilesystemBackend, route large tool results to /tmp to avoid polluting
    # the working directory. For sandbox backends, no special routing is needed.
    if sandbox is None:
        # Local mode: Route large results to a unique temp directory
        large_results_backend = FilesystemBackend(
            root_dir=tempfile.mkdtemp(prefix="deepagents_large_results_"),
            virtual_mode=True,
        )
        conversation_history_backend = FilesystemBackend(
            root_dir=tempfile.mkdtemp(prefix="deepagents_conversation_history_"),
            virtual_mode=True,
        )
        composite_backend = CompositeBackend(
            default=backend,
            routes={
                "/large_tool_results/": large_results_backend,
                "/conversation_history/": conversation_history_backend,
            },
        )
    else:
        # Sandbox mode: No special routing needed
        composite_backend = CompositeBackend(
            default=backend,
            routes={},
        )

    from deepagents.middleware.summarization import create_summarization_tool_middleware

    agent_middleware.append(
        create_summarization_tool_middleware(model, composite_backend)
    )

    # Create the agent
    all_subagents: list[SubAgent | CompiledSubAgent | AsyncSubAgent] = [
        *custom_subagents,
        *(async_subagents or []),
    ]
    agent = create_deep_agent(
        model=model,
        system_prompt=system_prompt,
        tools=tools,
        backend=composite_backend,
        middleware=agent_middleware,
        interrupt_on=interrupt_on,
        checkpointer=checkpointer,
        subagents=all_subagents or None,
    ).with_config(config)
    return agent, composite_backend



import textwrap
from pathlib import Path

from langchain_core.messages import ToolMessage

from deepagents_cli.file_ops import FileOpTracker, build_approval_preview


def test_tracker_records_read_lines(tmp_path: Path) -> None:
    tracker = FileOpTracker(assistant_id=None)
    path = tmp_path / "example.py"

    tracker.start_operation(
        "read_file",
        {"file_path": str(path), "offset": 0, "limit": 100},
        "read-1",
    )

    message = ToolMessage(
        content="    1\tline one\n    2\tline two\n",
        tool_call_id="read-1",
        name="read_file",
    )
    record = tracker.complete_with_message(message)

    assert record is not None
    assert record.metrics.lines_read == 2
    assert record.metrics.start_line == 1
    assert record.metrics.end_line == 2


def test_tracker_records_write_diff(tmp_path: Path) -> None:
    tracker = FileOpTracker(assistant_id=None)
    file_path = tmp_path / "created.txt"

    tracker.start_operation(
        "write_file",
        {"file_path": str(file_path)},
        "write-1",
    )

    file_path.write_text("hello world\nsecond line\n")

    message = ToolMessage(
        content=f"Updated file {file_path}",
        tool_call_id="write-1",
        name="write_file",
    )
    record = tracker.complete_with_message(message)

    assert record is not None
    assert record.metrics.lines_written == 2
    assert record.metrics.lines_added == 2
    assert record.diff is not None
    assert "+hello world" in record.diff


def test_tracker_records_edit_diff(tmp_path: Path) -> None:
    tracker = FileOpTracker(assistant_id=None)
    file_path = tmp_path / "functions.py"
    file_path.write_text(
        textwrap.dedent(
            """\
        def greet():
            return "hello"
        """
        )
    )

    tracker.start_operation(
        "edit_file",
        {"file_path": str(file_path)},
        "edit-1",
    )

    file_path.write_text(
        textwrap.dedent(
            """\
        def greet():
            return "hi"

        def wave():
            return "wave"
        """
        )
    )

    message = ToolMessage(
        content=f"Successfully replaced 1 instance(s) of the string in '{file_path}'",
        tool_call_id="edit-1",
        name="edit_file",
    )
    record = tracker.complete_with_message(message)

    assert record is not None
    assert record.metrics.lines_added >= 1
    assert record.metrics.lines_removed >= 1
    assert record.diff is not None
    assert '-    return "hello"' in record.diff
    assert '+    return "hi"' in record.diff


def test_build_approval_preview_generates_diff(tmp_path: Path) -> None:
    target = tmp_path / "notes.txt"
    target.write_text("alpha\nbeta\n")

    preview = build_approval_preview(
        "edit_file",
        {
            "file_path": str(target),
            "old_string": "beta",
            "new_string": "gamma",
            "replace_all": False,
        },
        assistant_id=None,
    )

    assert preview is not None
    assert preview.diff is not None
    assert "+gamma" in preview.diff



"""Unit tests for subagent loading functionality."""

from pathlib import Path

from deepagents_cli.subagents import (
    _load_subagents_from_dir,
    _parse_subagent_file,
    list_subagents,
)


def make_subagent_content(
    name: str,
    description: str,
    model: str | None = None,
    system_prompt: str | None = None,
) -> str:
    """Create subagent markdown content with YAML frontmatter."""
    model_line = f"model: {model}\n" if model else ""
    prompt = (
        system_prompt
        or f"You are a {name} assistant.\n\n## Instructions\nDo your job well."
    )
    return f"""---
name: {name}
description: {description}
{model_line}---

{prompt}
"""


class TestParseSubagentFile:
    """Test _parse_subagent_file function."""

    def test_parse_valid_subagent_with_all_fields(self, tmp_path: Path) -> None:
        """Test parsing a valid subagent file with all fields."""
        subagent_file = tmp_path / "researcher.md"
        subagent_file.write_text(
            make_subagent_content(
                "researcher",
                "Research topics on the web",
                model="anthropic:claude-haiku-4-5-20251001",
            )
        )

        result = _parse_subagent_file(subagent_file)

        assert result is not None
        assert result["name"] == "researcher"
        assert result["description"] == "Research topics on the web"
        assert result["model"] == "anthropic:claude-haiku-4-5-20251001"
        assert "researcher assistant" in result["system_prompt"]
        assert "## Instructions" in result["system_prompt"]
        assert result["path"] == str(subagent_file)

    def test_parse_subagent_without_model(self, tmp_path: Path) -> None:
        """Test parsing subagent without optional model field."""
        subagent_file = tmp_path / "helper.md"
        subagent_file.write_text(make_subagent_content("helper", "A helpful assistant"))

        result = _parse_subagent_file(subagent_file)

        assert result is not None
        assert result["name"] == "helper"
        assert result["description"] == "A helpful assistant"
        assert result["model"] is None

    def test_parse_subagent_with_multiline_system_prompt(self, tmp_path: Path) -> None:
        """Test parsing subagent with complex multiline system prompt."""
        subagent_file = tmp_path / "writer.md"
        content = """---
name: writer
description: Write content
---

You are a skilled writer.

## Guidelines

1. Write clearly
2. Use proper grammar
3. Be concise

## Output Format

Always structure your response with headings.
"""
        subagent_file.write_text(content)

        result = _parse_subagent_file(subagent_file)

        assert result is not None
        assert "## Guidelines" in result["system_prompt"]
        assert "## Output Format" in result["system_prompt"]
        assert "1. Write clearly" in result["system_prompt"]

    def test_parse_subagent_missing_name(self, tmp_path: Path) -> None:
        """Test that subagent without name is rejected."""
        subagent_file = tmp_path / "invalid.md"
        subagent_file.write_text("""---
description: Missing name field
---

Content
""")

        assert _parse_subagent_file(subagent_file) is None

    def test_parse_subagent_missing_description(self, tmp_path: Path) -> None:
        """Test that subagent without description is rejected."""
        subagent_file = tmp_path / "invalid.md"
        subagent_file.write_text("""---
name: invalid
---

Content
""")

        assert _parse_subagent_file(subagent_file) is None

    def test_parse_subagent_no_frontmatter(self, tmp_path: Path) -> None:
        """Test that file without frontmatter is rejected."""
        subagent_file = tmp_path / "invalid.md"
        subagent_file.write_text("# Just markdown\n\nNo frontmatter.")

        assert _parse_subagent_file(subagent_file) is None

    def test_parse_subagent_invalid_yaml(self, tmp_path: Path) -> None:
        """Test that invalid YAML is rejected."""
        subagent_file = tmp_path / "invalid.md"
        subagent_file.write_text("""---
name: [unclosed
description: test
---

Content
""")

        assert _parse_subagent_file(subagent_file) is None

    def test_parse_subagent_empty_name(self, tmp_path: Path) -> None:
        """Test that empty name is rejected."""
        subagent_file = tmp_path / "invalid.md"
        subagent_file.write_text("""---
name: ""
description: Has empty name
---

Content
""")

        assert _parse_subagent_file(subagent_file) is None

    def test_parse_subagent_non_string_name(self, tmp_path: Path) -> None:
        """Test that non-string name is rejected."""
        subagent_file = tmp_path / "invalid.md"
        subagent_file.write_text("""---
name: 123
description: Has numeric name
---

Content
""")

        assert _parse_subagent_file(subagent_file) is None

    def test_parse_subagent_non_string_model(self, tmp_path: Path) -> None:
        """Test that non-string model is rejected."""
        subagent_file = tmp_path / "invalid.md"
        subagent_file.write_text("""---
name: test
description: Test
model: 123
---

Content
""")

        assert _parse_subagent_file(subagent_file) is None

    def test_parse_subagent_nonexistent_file(self, tmp_path: Path) -> None:
        """Test that nonexistent file returns None."""
        subagent_file = tmp_path / "nonexistent.md"

        assert _parse_subagent_file(subagent_file) is None

    def test_parse_subagent_frontmatter_not_dict(self, tmp_path: Path) -> None:
        """Test that non-dict frontmatter is rejected."""
        subagent_file = tmp_path / "invalid.md"
        subagent_file.write_text("""---
- item1
- item2
---

Content
""")

        assert _parse_subagent_file(subagent_file) is None


class TestLoadSubagentsFromDir:
    """Test _load_subagents_from_dir function."""

    def test_load_from_empty_directory(self, tmp_path: Path) -> None:
        """Test loading from empty directory."""
        agents_dir = tmp_path / "agents"
        agents_dir.mkdir()

        result = _load_subagents_from_dir(agents_dir, "user")
        assert result == {}

    def test_load_single_subagent(self, tmp_path: Path) -> None:
        """Test loading a single subagent."""
        agents_dir = tmp_path / "agents"
        folder = agents_dir / "researcher"
        folder.mkdir(parents=True)
        (folder / "AGENTS.md").write_text(
            make_subagent_content("researcher", "Research assistant")
        )

        result = _load_subagents_from_dir(agents_dir, "user")

        assert len(result) == 1
        assert "researcher" in result
        assert result["researcher"]["source"] == "user"

    def test_load_multiple_subagents(self, tmp_path: Path) -> None:
        """Test loading multiple subagents."""
        agents_dir = tmp_path / "agents"

        for name in ["researcher", "writer", "reviewer"]:
            folder = agents_dir / name
            folder.mkdir(parents=True)
            (folder / "AGENTS.md").write_text(
                make_subagent_content(name, f"{name.title()} assistant")
            )

        result = _load_subagents_from_dir(agents_dir, "project")

        assert len(result) == 3
        assert all(s["source"] == "project" for s in result.values())

    def test_load_skips_misnamed_files(self, tmp_path: Path) -> None:
        """Test that files not matching expected name are skipped."""
        agents_dir = tmp_path / "agents"
        folder = agents_dir / "researcher"
        folder.mkdir(parents=True)
        # Wrong filename - should be AGENTS.md
        (folder / "agent.md").write_text(
            make_subagent_content("researcher", "Research assistant")
        )

        result = _load_subagents_from_dir(agents_dir, "user")
        assert result == {}

    def test_load_skips_invalid_subagents(self, tmp_path: Path) -> None:
        """Test that invalid subagents are skipped."""
        agents_dir = tmp_path / "agents"

        # Valid subagent
        valid_folder = agents_dir / "valid"
        valid_folder.mkdir(parents=True)
        (valid_folder / "AGENTS.md").write_text(
            make_subagent_content("valid", "Valid assistant")
        )

        # Invalid subagent (missing description)
        invalid_folder = agents_dir / "invalid"
        invalid_folder.mkdir(parents=True)
        (invalid_folder / "AGENTS.md").write_text("""---
name: invalid
---
Content
""")

        result = _load_subagents_from_dir(agents_dir, "user")

        assert len(result) == 1
        assert "valid" in result

    def test_load_skips_files_in_root(self, tmp_path: Path) -> None:
        """Test that files directly in agents dir (not in subfolders) are skipped."""
        agents_dir = tmp_path / "agents"
        agents_dir.mkdir()
        (agents_dir / "stray.md").write_text(
            make_subagent_content("stray", "Stray file")
        )

        result = _load_subagents_from_dir(agents_dir, "user")
        assert result == {}

    def test_load_nonexistent_directory(self, tmp_path: Path) -> None:
        """Test loading from nonexistent directory."""
        result = _load_subagents_from_dir(tmp_path / "nonexistent", "user")
        assert result == {}


class TestListSubagents:
    """Test list_subagents function."""

    def test_list_no_directories(self) -> None:
        """Test listing with no directories specified."""
        result = list_subagents()
        assert result == []

    def test_list_user_only(self, tmp_path: Path) -> None:
        """Test listing from user directory only."""
        user_dir = tmp_path / "user_agents"
        folder = user_dir / "researcher"
        folder.mkdir(parents=True)
        (folder / "AGENTS.md").write_text(
            make_subagent_content("researcher", "Research assistant")
        )

        result = list_subagents(user_agents_dir=user_dir)

        assert len(result) == 1
        assert result[0]["name"] == "researcher"
        assert result[0]["source"] == "user"

    def test_list_project_only(self, tmp_path: Path) -> None:
        """Test listing from project directory only."""
        project_dir = tmp_path / "project_agents"
        folder = project_dir / "reviewer"
        folder.mkdir(parents=True)
        (folder / "AGENTS.md").write_text(
            make_subagent_content("reviewer", "Code reviewer")
        )

        result = list_subagents(project_agents_dir=project_dir)

        assert len(result) == 1
        assert result[0]["name"] == "reviewer"
        assert result[0]["source"] == "project"

    def test_list_both_sources(self, tmp_path: Path) -> None:
        """Test listing from both user and project directories."""
        user_dir = tmp_path / "user_agents"
        project_dir = tmp_path / "project_agents"

        # User subagent
        user_folder = user_dir / "researcher"
        user_folder.mkdir(parents=True)
        (user_folder / "AGENTS.md").write_text(
            make_subagent_content("researcher", "Research assistant")
        )

        # Project subagent
        project_folder = project_dir / "reviewer"
        project_folder.mkdir(parents=True)
        (project_folder / "AGENTS.md").write_text(
            make_subagent_content("reviewer", "Code reviewer")
        )

        result = list_subagents(
            user_agents_dir=user_dir,
            project_agents_dir=project_dir,
        )

        assert len(result) == 2
        names = {s["name"] for s in result}
        assert names == {"researcher", "reviewer"}

    def test_list_project_overrides_user(self, tmp_path: Path) -> None:
        """Test that project subagents override user subagents with same name."""
        user_dir = tmp_path / "user_agents"
        project_dir = tmp_path / "project_agents"

        # User version
        user_folder = user_dir / "shared"
        user_folder.mkdir(parents=True)
        (user_folder / "AGENTS.md").write_text(
            make_subagent_content("shared", "User version")
        )

        # Project version (same name)
        project_folder = project_dir / "shared"
        project_folder.mkdir(parents=True)
        (project_folder / "AGENTS.md").write_text(
            make_subagent_content("shared", "Project version")
        )

        result = list_subagents(
            user_agents_dir=user_dir,
            project_agents_dir=project_dir,
        )

        assert len(result) == 1
        assert result[0]["name"] == "shared"
        assert result[0]["description"] == "Project version"
        assert result[0]["source"] == "project"

    def test_list_empty_directories(self, tmp_path: Path) -> None:
        """Test listing from empty directories."""
        user_dir = tmp_path / "user_agents"
        project_dir = tmp_path / "project_agents"
        user_dir.mkdir()
        project_dir.mkdir()

        result = list_subagents(
            user_agents_dir=user_dir,
            project_agents_dir=project_dir,
        )
        assert result == []

    def test_list_nonexistent_directories(self, tmp_path: Path) -> None:
        """Test listing from nonexistent directories."""
        result = list_subagents(
            user_agents_dir=tmp_path / "nonexistent_user",
            project_agents_dir=tmp_path / "nonexistent_project",
        )
        assert result == []

    def test_list_with_model_field(self, tmp_path: Path) -> None:
        """Test that model field is correctly loaded."""
        user_dir = tmp_path / "agents"
        folder = user_dir / "fast-researcher"
        folder.mkdir(parents=True)
        (folder / "AGENTS.md").write_text(
            make_subagent_content(
                "fast-researcher",
                "Fast research using Haiku",
                model="anthropic:claude-haiku-4-5-20251001",
            )
        )

        result = list_subagents(user_agents_dir=user_dir)

        assert len(result) == 1
        assert result[0]["model"] == "anthropic:claude-haiku-4-5-20251001"
